import numpy
import math

from .ctng import constructive_neuronal_geometry
from .graphicsPrimitives import (
    Sphere,
    Cone,
    Cylinder,
    SkewCone,
    Plane,
    Union,
    Intersection,
    SphereCone,
)
from .GeneralizedVoxelization import voxelize
from .simplevolume_helper import simplevolume
from .surface_a import surface_area
from ..options import ics_distance_threshold
from ..rxdException import RxDException
import warnings
from neuron import h, _sec_db


def find_parent_seg(join, sdict, objects):

    if not join:
        return None
    elif join[0] not in objects:
        pseg = sdict[
            (
                join[0]._x0,
                join[0]._y0,
                join[0]._z0,
                join[0]._x1,
                join[0]._y1,
                join[0]._z1,
            )
        ]
        # better be all in same cell; so just set root once
        h.distance(0, h.SectionRef(sec=pseg.sec).root(0))
        closest = h.distance(pseg)

    # any other possible instance?

    for item in join:
        if item not in objects:
            s = sdict[(item._x0, item._y0, item._z0, item._x1, item._y1, item._z1)]
            d = h.distance(s)
            if d < closest:
                pseg = s
                closest = d

    return pseg


def all_in(dist):
    for i in dist:
        if i > ics_distance_threshold:
            return False
    return True


def sort_spheres_last(item):
    return 1 if isinstance(item, Sphere) else 0


def fullmorph(source, dx, soma_step=100, mesh_grid=None, relevant_pts=None):

    """Input: object source; arguments to pass to ctng
    Output: all voxels with SA and volume associated, categorized by segment"""
    source = list(source)
    morphology = constructive_neuronal_geometry(
        source, soma_step, dx, relevant_pts=relevant_pts
    )
    (
        join_objects,
        cones,
        segment_dict,
        join_groups,
        object_pts,
        soma_objects,
    ) = morphology

    # grid setup
    if mesh_grid:
        grid = mesh_grid
        if grid["dx"] != dx or grid["dy"] != dx or grid["dz"] != dx:
            raise RxDException(
                "Error: Inconsistent vocalization. Mesh grid voxels (%g, %g, %g) differs from requested dx (%g, %g %g),",
                (grid["dx"], grid["dy"], grid["dz"], dx, dx, dx),
            )
    else:

        def minmax(lst, old):
            return (
                [min(min(lst), old[0]), max(max(lst), old[1])]
                if old != []
                else [min(lst), max(lst)]
            )

        xs, ys, zs, diams, soma_idx = [], [], [], [], []
        arcs = float("inf")
        for i, sec in enumerate(source):
            if relevant_pts:
                rng = relevant_pts[i]
            else:
                rng = range(sec.n3d())
            if sec.hoc_internal_name() in _sec_db:
                soma_idx.append(i)
            else:
                xs = minmax([sec.x3d(i) for i in rng], xs)
                ys = minmax([sec.y3d(i) for i in rng], ys)
                zs = minmax([sec.z3d(i) for i in rng], zs)
                diams = minmax([sec.diam3d(i) for i in rng if sec.diam3d(i) > 0], diams)
                arcs = min(
                    [
                        sec.arc3d(i + 1) - sec.arc3d(i)
                        for i in rng[:-1]
                        if sec.arc3d(i + 1) > sec.arc3d(i)
                    ]
                    + [arcs]
                )

        # TODO: include segment boundaries when checking cone lengths
        # warning on minimum size of dx, only considering positive lengths
        if diams:
            check = min(diams[0] / math.sqrt(3), arcs / math.sqrt(3))
            if dx > check:
                warnings.warn(
                    "Resolution may be too low. To guarantee accurate voxelization, use a dx <= {}.".format(
                        check
                    )
                )

        for idx in soma_idx:
            sec = source[idx]
            rng = relevant_pts[idx] if relevant_pts else range(sec.n3d())
            xs = minmax([sec.x3d(i) for i in rng], xs)
            ys = minmax([sec.y3d(i) for i in rng], ys)
            zs = minmax([sec.z3d(i) for i in rng], zs)
            diams = minmax([sec.diam3d(i) for i in rng if sec.diam3d(i) > 0], diams)

        dy = dz = dx  # ever going to change this?

        margin = diams[1] + 2 * dx
        grid = {
            "xlo": xs[0] - margin,
            "xhi": xs[1] + margin,
            "ylo": ys[0] - margin,
            "yhi": ys[1] + margin,
            "zlo": zs[0] - margin,
            "zhi": zs[1] + margin,
            "dx": dx,
            "dy": dy,
            "dz": dz,
        }
    ##########################################################
    final_seg_dict = {}

    # soma: modified when ctng properly assigns soma obj to segments
    if soma_objects:
        for item, seg in soma_objects.items():
            if seg in final_seg_dict:
                final_seg_dict[seg].append(item)
            else:
                final_seg_dict[seg] = [item]

    # assign join objects
    for jg in join_groups:
        seg = find_parent_seg(jg, segment_dict, join_objects)
        for item in jg:
            # if isinstance(item, Sphere): continue # TODO: don't do this... just a test
            if (not (isinstance(item, Cone) or isinstance(item, Cylinder))) or (
                item in join_objects
            ):
                if seg in final_seg_dict.keys():
                    final_seg_dict[seg].append(item)
                else:
                    final_seg_dict[seg] = [item]

    # complete final segment dictionary
    for cone in cones:
        seg = segment_dict[(cone._x0, cone._y0, cone._z0, cone._x1, cone._y1, cone._z1)]
        if seg in final_seg_dict.keys():
            final_seg_dict[seg].append(cone)
        else:
            final_seg_dict[seg] = [cone]

    # voxelize all the objects and assign voxels
    # output dictionaries of internal and surface voxels
    # assign voxels to segments
    missed = 0
    missed_voxels = set()
    total_surface_voxels = {}
    final_intern_voxels = {}  # final output of internal voxels

    for do_spheres in [False, True]:
        for seg in final_seg_dict:
            distance_root = h.SectionRef(sec=seg.sec).root(0)
            for item in [
                my_item
                for my_item in final_seg_dict[seg]
                if isinstance(my_item, Sphere) == do_spheres
            ]:
                if item in object_pts:
                    [yesvox, surface, miss] = voxelize(grid, item, object_pts[item])
                else:
                    [yesvox, surface, miss] = voxelize(grid, item)
                if miss:
                    missed += 1
                    missed_voxels.add(miss)

                # must take only the internal voxels for that item (set diff)
                yesvox = yesvox - set(surface.keys())
                for i in yesvox:
                    if i in final_intern_voxels.keys():
                        if h.distance(distance_root, seg) < h.distance(
                            distance_root, final_intern_voxels[i][1]
                        ) and not isinstance(item, Sphere):
                            final_intern_voxels[i][1] = seg
                    else:
                        final_intern_voxels[i] = [dx**3, seg]

                for i in surface.keys():
                    if i in total_surface_voxels.keys():
                        total_surface_voxels[i][0].append(item)
                        # update the distances list to the minimum at each vertex
                        total_surface_voxels[i][1] = [
                            min(total_surface_voxels[i][1][j], surface[i][j])
                            for j in range(8)
                        ]
                        if h.distance(distance_root, seg) < h.distance(
                            distance_root, total_surface_voxels[i][2]
                        ) and not isinstance(item, Sphere):
                            total_surface_voxels[i][2] = seg
                    else:
                        total_surface_voxels[i] = [[item], surface[i], seg]

    # take internal voxels out of surface voxels
    for vox in final_intern_voxels.keys():
        if vox in total_surface_voxels.keys():
            del total_surface_voxels[vox]

    # calculate volume and SA for surface voxels, taking into account multiple item crossovers
    final_surface_voxels = {}
    for vox in total_surface_voxels.keys():
        # modified version of SV1 functions verts_in that returns volume and distances
        [itemlist, distances, seg] = total_surface_voxels[vox]
        # check if the surface voxel is actually internal after updating all vertex distances:
        if all_in(distances):
            final_intern_voxels[vox] = [dx**3, seg]
            # no need to delete from surface voxels, it is never added to final surface voxels
        else:
            V = simplevolume(itemlist, distances, vox, grid)
            A = surface_area(itemlist, vox, grid)
            # if the voxel form a corner of itemlist it may have no surface
            # area
            if A == 0:
                final_intern_voxels[vox] = [dx**3, seg]
            else:
                final_surface_voxels[vox] = [V, A, seg]

    def has_vox(*vox):
        return vox in final_surface_voxels or vox in final_intern_voxels

    # if an internal voxel doesn't have a neighbor... then it's not internal
    # TODO: this should never happen, and seems to only happen when there
    #       is surface that passes exactly through the boundary
    #       Figure out how to handle that, and remove this; it slows things down a lot
    actually_surface_voxels = []
    for vox, (vol, seg) in final_intern_voxels.items():
        i, j, k = vox
        area = 0
        if not has_vox(i + 1, j, k):
            area += dx**2
        if not has_vox(i - 1, j, k):
            area += dx**2
        if not has_vox(i, j + 1, k):
            area += dx**2
        if not has_vox(i, j - 1, k):
            area += dx**2
        if not has_vox(i, j, k + 1):
            area += dx**2
        if not has_vox(i, j, k - 1):
            area += dx**2
        if area:
            final_surface_voxels[vox] = [vol, area, seg]
            actually_surface_voxels.append(vox)

    for vox in actually_surface_voxels:
        del final_intern_voxels[vox]

    # make sure to keep track of poss_missed for each cone; total should be 0
    """if missed > 0:
        print("{} objects have inaccurate voxelization, probably due to resolution errors.".format(missed))
        print("Missed voxels total: ", len(missed_voxels))
    """
    return final_intern_voxels, final_surface_voxels, grid



from neuron import h
from neuron.expect_hocerr import expect_err


def test_template_err():

    h(
        """
begintemplate TestTemplateErr1
proc init() {
    create soma
}
endtemplate TestTemplateErr1
    """
    )

    expect_err("m = h.TestTemplateErr1()")

    h(
        """
begintemplate TestTemplateErr2
proc init() {
}
proc setfoo() {
    foo = 5 // 0 default if not called instead of UNDEF
}
endtemplate TestTemplateErr2
    """
    )

    m = h.TestTemplateErr2()
    expect_err('h.execute("create foo", m)')
    assert m.foo == 0.0


def test_template_err2():
    h(
        """
proc test_template_err2() {
    create test_template_err_sec
}
    """
    )
    h.test_template_err2()
    s = h.test_template_err_sec
    s.nseg = 3
    s.insert("hh")
    h.topology()
    print(s.psection())
    h.delete_section(sec=h.test_template_err_sec)


if __name__ == "__main__":
    test_template_err()
    test_template_err2()



from .xml2nrn import *


def group(self, node):
    groupName = str(node.text.strip())
    self.biomechs_[-1].parms_[-1].group_index_ = self.groupname2index_[groupName]
    if debug:
        print("Content: ", groupName, " relevant to biogroup: ", self.biomechs_[-1])


def mechanism(self, node):
    name = str(node.get("name"))
    pc = node.get("passiveConductance")
    if pc is not None:
        if pc == "true" or pc == "1":
            print(
                "Substituting passive conductance",
                name,
                " in file for inbuilt mechanism pas as attribute passiveConductance = true ",
            )
            name = "pas"
    self.biomechs_.append(BioMech(name))


def mechanism_end(self, node):
    m = self.biomechs_[-1]
    for p in m.parms_:
        cg = self.cablegroups_[p.group_index_]
        if len(cg.mechs_) == 0 or cg.mechs_[-1].name_ != m.name_:
            cg.mechs_.append(BioMech(m.name_))
        cg.mechs_[-1].parms_.append(p)


def parameter(self, node):
    value = float(node.get("value"))
    name = node.get("name")
    if name is not None:
        # convert from physiological to NEURON units
        if self.is_physiological_units_:
            if name == "gmax":
                value *= 0.001
        # NEURON ModelView converts g_pas to gmax.
        # following does only that one
        name = name.encode("ascii") + "_" + self.biomechs_[-1].name_
        if name == "gmax_pas":
            name = "g_pas"
        self.biomechs_[-1].parms_.append(BioParm(name, value))
    else:
        if self.biomechs_[-1].name_ == "Ra":
            value *= 1000.0
        self.biomechs_[-1].parms_.append(BioParm(self.biomechs_[-1].name_, value))


def spec_axial_resistance(self, node):
    self.biomechs_.append(BioMech("Ra"))


def spec_capacitance(self, node):
    self.biomechs_.append(BioMech("cm"))



# Example copied from the RxD tutorial
# http://www.neuron.yale.edu/neuron/static/docs/rxd/index.html
from neuron import rxd, h, gui
import numpy

sec = h.Section()
sec.L = 100
sec.diam = 1
sec.nseg = 100

h.CVode().active(True)
h.CVode().atol(1e-4)


caDiff = 0.016
ip3Diff = 0.283
cac_init = 1.0e-4
ip3_init = 0.1
gip3r = 12040
gserca = 0.3913
gleak = 6.020
kserca = 0.1
kip3 = 0.15
kact = 0.4
ip3rtau = 2000.0

# These parameters where missing in the tutorial so arbitrary values were chosen
# any resemblance to experimental values is purely coincidental.
fc = 0.7
fe = 0.3
caCYT_init = 0.1


cyt = rxd.Region(
    h.allsec(),
    name="cyt",
    nrn_region="i",
    geometry=rxd.FractionalVolume(fc, surface_fraction=1),
)

er = rxd.Region(h.allsec(), name="er", geometry=rxd.FractionalVolume(fe / 2.0))

cyt_er_membrane = rxd.Region(
    h.allsec(), name="mem", geometry=rxd.ScalableBorder(1, on_cell_surface=False)
)

ca = rxd.Species([cyt, er], d=caDiff, name="ca", charge=2, initial=caCYT_init)


ip3 = rxd.Species(cyt, d=ip3Diff, name="ip3", initial=ip3_init)
ip3r_gate_state = rxd.Species(cyt_er_membrane, name="gate", initial=0.8)
h_gate = ip3r_gate_state[cyt_er_membrane]
minf = ip3[cyt] * 1000.0 * ca[cyt] / (ip3[cyt] + kip3) / (1000.0 * ca[cyt] + kact)
k = gip3r * (minf * h_gate) ** 3

serca = rxd.MultiCompartmentReaction(
    ca[cyt],
    ca[er],
    gserca / ((kserca / (1000.0 * ca[cyt])) ** 2 + 1),
    membrane=cyt_er_membrane,
    custom_dynamics=True,
)
leak = rxd.MultiCompartmentReaction(
    ca[er], ca[cyt], gleak, gleak, membrane=cyt_er_membrane
)


ip3r = rxd.MultiCompartmentReaction(ca[er], ca[cyt], k, k, membrane=cyt_er_membrane)
ip3rg = rxd.Rate(h_gate, (1.0 / (1 + 1000.0 * ca[cyt] / (0.3)) - h_gate) / ip3rtau)

cacyt_trace = h.Vector()
cacyt_trace.record(ca[cyt].nodes(sec)(0.5)[0]._ref_concentration)

caer_trace = h.Vector()
caer_trace.record(ca[er].nodes(sec)(0.5)[0]._ref_concentration)

ip3_trace = h.Vector()
ip3_trace.record(ip3.nodes(sec)(0.5)[0]._ref_concentration)

times = h.Vector()
times.record(h._ref_t)

h.finitialize(-65)

cae_init = (0.0017 - cac_init * fc) / fe
ca[er].concentration = cae_init

for node in ip3.nodes:
    if node.x < 0.2:
        node.concentration = 2


h.CVode().re_init()
h.continuerun(1000)

if __name__ == "__main__":
    from matplotlib import pyplot

    pyplot.plot(times, cacyt_trace, label="ca[cyt]")
    pyplot.plot(times, caer_trace, label="ca[er]")
    pyplot.plot(times, ip3_trace, label="ip3")
    pyplot.legend()
    pyplot.show()



import sys
from neuron import h

h.nrnmpi_init()
pc = h.ParallelContext()

id = int(pc.id())
nhost = int(pc.nhost())

if "--expected-hosts" in sys.argv:
    expected_nhost_index = sys.argv.index("--expected-hosts") + 1
    expected_nhost = int(sys.argv[expected_nhost_index])
else:
    # if there is no cli arg then accept nhost as expected host count
    expected_nhost = nhost

print("I am %d of %d" % (id, nhost))

if nhost != expected_nhost:
    sys.exit(
        "MPI Error : #ranks ({}) different that the expected ({})".format(
            nhost, expected_nhost
        )
    )

pc.barrier()
h.quit()



def test_savestate_basic(neuron_nosave_instance):
    """Test SaveState works rxd when rxd states are not saved/restored by save
    state"""
    h, rxd, save_path = neuron_nosave_instance
    vtrap, v = rxd.rxdmath.vtrap, rxd.v
    exp, log = rxd.rxdmath.exp, rxd.rxdmath.log

    def saveSS():
        svst = h.SaveState()
        svst.save()
        svst_extra = save_rxd_extra()  # only for below model with one segment.
        return svst, svst_extra

    def restoreSS(svst, svst_extra):
        svst.restore()
        restore_rxd_extra(svst_extra)  # only for below model with one segment.
        h.fcurrent()  # failed hope was that this would properly initialize the currents.

    # parameters
    h.celsius = 6.3
    e = 1.60217662e-19
    scale = 1e-14 / e
    gnabar = 0.12 * scale  # molecules/um2 ms mV
    gkbar = 0.036 * scale
    gl = 0.0003 * scale
    el = -54.3
    q10 = 3.0 ** ((h.celsius - 6.3) / 10.0)

    # sodium activation 'm'
    alpha = 0.1 * vtrap(-(v + 40.0), 10)
    beta = 4.0 * exp(-(v + 65) / 18.0)
    mtau = 1.0 / (q10 * (alpha + beta))
    minf = alpha / (alpha + beta)

    # sodium inactivation 'h'
    alpha = 0.07 * exp(-(v + 65.0) / 20.0)
    beta = 1.0 / (exp(-(v + 35.0) / 10.0) + 1.0)
    htau = 1.0 / (q10 * (alpha + beta))
    hinf = alpha / (alpha + beta)

    # potassium activation 'n'
    alpha = 0.01 * vtrap(-(v + 55.0), 10.0)
    beta = 0.125 * exp(-(v + 65.0) / 80.0)
    ntau = 1.0 / (q10 * (alpha + beta))
    ninf = alpha / (alpha + beta)

    somaA = h.Section("somaA")
    somaA.pt3dclear()
    somaA.pt3dadd(-90, 0, 0, 30)
    somaA.pt3dadd(-60, 0, 0, 30)
    somaA.nseg = 1  # 11

    # Where?
    # intracellular
    cyt = rxd.Region(h.allsec(), name="cyt", nrn_region="i")

    # membrane
    mem = rxd.Region(h.allsec(), name="cell_mem", geometry=rxd.membrane())

    # extracellular
    ecs = rxd.Extracellular(-100, -100, -100, 100, 100, 100, dx=33)

    # Who?  ions & gates

    # intracellular sodium & potassium
    na = rxd.Species([cyt, mem], name="na", d=1, charge=1, initial=10)
    k = rxd.Species([cyt, mem], name="k", d=1, charge=1, initial=54.4)

    # extracellular parameters provide a constant concentration for the Nernst potential and reactions.
    kecs = rxd.Parameter(ecs, name="k", charge=1, value=2.5)
    naecs = rxd.Parameter(ecs, name="na", charge=1, value=140)

    # an undistinguished charged ion for the leak current
    x = rxd.Species([cyt, mem, ecs], name="x", charge=1)

    # define the various species and parameters on the intracellular and extracellular regions
    ki, ko, nai, nao, xi, xo = k[cyt], kecs[ecs], na[cyt], naecs[ecs], x[cyt], x[ecs]

    # the gating states
    ngate = rxd.State([cyt, mem], name="ngate", initial=0.24458654944007166)
    mgate = rxd.State([cyt, mem], name="mgate", initial=0.028905534475191907)
    hgate = rxd.State([cyt, mem], name="hgate", initial=0.7540796658225246)

    # parameter to limit rxd reaction to somaA
    paramA = rxd.Parameter(
        [cyt, mem], name="paramA", value=lambda nd: 1 if nd.segment in somaA else 0
    )

    # What? gates and currents
    m_gate = rxd.Rate(mgate, (minf - mgate) / mtau)
    h_gate = rxd.Rate(hgate, (hinf - hgate) / htau)
    n_gate = rxd.Rate(ngate, (ninf - ngate) / ntau)

    # Nernst potentials
    ena = 1e3 * h.R * (h.celsius + 273.15) * log(nao / nai) / h.FARADAY
    ek = 1e3 * h.R * (h.celsius + 273.15) * log(ko / ki) / h.FARADAY

    gna = paramA * gnabar * mgate**3 * hgate
    gk = paramA * gkbar * ngate**4

    na_current = rxd.MultiCompartmentReaction(
        nai, nao, gna * (v - ena), mass_action=False, membrane=mem, membrane_flux=True
    )
    k_current = rxd.MultiCompartmentReaction(
        ki, ko, gk * (v - ek), mass_action=False, membrane=mem, membrane_flux=True
    )
    leak_current = rxd.MultiCompartmentReaction(
        xi,
        xo,
        paramA * gl * (v - el),
        mass_action=False,
        membrane=mem,
        membrane_flux=True,
    )

    def save_rxd_extra():
        # only for one segment
        sav_rxd_extra_ = []
        sav_rxd_extra_.append(mgate[cyt].nodes(somaA(0.5)).value)
        sav_rxd_extra_.append(ngate[cyt].nodes(somaA(0.5)).value)
        sav_rxd_extra_.append(hgate[cyt].nodes(somaA(0.5)).value)
        sav_rxd_extra_.append(mgate[mem].nodes(somaA(0.5)).value)
        sav_rxd_extra_.append(ngate[mem].nodes(somaA(0.5)).value)
        sav_rxd_extra_.append(hgate[mem].nodes(somaA(0.5)).value)
        return sav_rxd_extra_

    def restore_rxd_extra(sav_rxd_extra_):
        mgate[cyt].nodes(somaA(0.5)).value = sav_rxd_extra_[0]
        ngate[cyt].nodes(somaA(0.5)).value = sav_rxd_extra_[1]
        hgate[cyt].nodes(somaA(0.5)).value = sav_rxd_extra_[2]
        mgate[mem].nodes(somaA(0.5)).value = sav_rxd_extra_[3]
        ngate[mem].nodes(somaA(0.5)).value = sav_rxd_extra_[4]
        hgate[mem].nodes(somaA(0.5)).value = sav_rxd_extra_[5]
        k[cyt].nodes.value = somaA(0.5).ki
        na[cyt].nodes.value = somaA(0.5).nai

    # return list of current values
    def curval():
        a = [h.t, somaA(0.5).v]
        a.append(mgate[cyt].nodes(somaA(0.5)).value)
        a.append(ngate[cyt].nodes(somaA(0.5)).value)
        a.append(hgate[cyt].nodes(somaA(0.5)).value)
        a.append(somaA(0.5).ik)
        a.append(somaA(0.5).ina)
        return a

    def test_savestate_rxd():
        t1 = 7
        t2 = 10
        # record
        tvec = h.Vector().record(h._ref_t)
        vvecA = h.Vector().record(somaA(0.5)._ref_v)
        mvecA = h.Vector().record(mgate[cyt].nodes(somaA(0.5))._ref_value)
        nvecA = h.Vector().record(ngate[cyt].nodes(somaA(0.5))._ref_value)
        hvecA = h.Vector().record(hgate[cyt].nodes(somaA(0.5))._ref_value)
        kvecA = h.Vector().record(somaA(0.5)._ref_ik)
        navecA = h.Vector().record(somaA(0.5)._ref_ina)

        vecs = [tvec, vvecA, mvecA, nvecA, hvecA, kvecA, navecA]

        # Full run to 100 to get standard results.
        h.finitialize(-70)
        h.continuerun(t2)
        vecs_std = [i.c() for i in vecs]

        # Run to t1, savestate, continue run to t2.
        # compare results to standard results (i.e. savestate should
        # not corrupt the run.)
        h.finitialize(-70)
        h.continuerun(t1)
        sav_val = curval()
        print(sav_val)
        svst, svst_extra = saveSS()
        print(curval())
        h.continuerun(t2)
        for i, vec in enumerate(vecs):
            if not vec.eq(vecs_std[i]):
                print(
                    "vecs[%d] error %g" % (i, vecs[i].c().sub(vecs_std[i]).abs().sum())
                )
        for i, vec in enumerate(vecs):
            assert vec.eq(vecs_std[i])

        # Another run stopping at t1, resizing the record vectors to 0
        # and continuing to t2 to be the standard for a later run that will be
        # savestate restored at t1.
        h.finitialize(-70)
        h.continuerun(t1)
        print(curval())
        for vec in vecs:
            vec.resize(0)
        h.continuerun(t2)
        vecs_std_save = [i.c() for i in vecs]

        # Finally, savestate restore at t1, continue to t2, and
        # compare vecs to vecs_std_save
        h.finitialize(-70)
        print(curval())
        restoreSS(svst, svst_extra)
        print(curval())
        for vec in vecs:
            vec.resize(0)
        h.continuerun(t2)
        for i, vec in enumerate(vecs):
            if not vec.eq(vecs_std_save[i]):
                print(
                    "vecs[%d] error %g"
                    % (i, vecs[i].c().sub(vecs_std_save[i]).abs().sum())
                )
        for i, vec in enumerate(vecs):
            assert vec.eq(vecs_std_save[i])
            pass

        h.topology()
        return vecs, vecs_std_save

    # run the test
    return test_savestate_rxd()


if __name__ == "__main__":
    from neuron import h, rxd

    h.load_file("stdrun.hoc")
    instance = h, rxd, None
    vecs, vecs_std_save = test_savestate_basic(instance)



from neuron import h, crxd as rxd

h.load_file("stdrun.hoc")

cell = h.Section(name="cell")
r = rxd.Region([cell])


def Declare_Parameters(**kwargs):
    """helper function enabling clean declaration of parameters in top namespace"""
    for key, value in kwargs.items():
        globals()[key] = rxd.Parameter(r, name=key, initial=value)


def Declare_Species(**kwargs):
    """helper function enabling clean declaration of species in top namespace"""
    for key, value in kwargs.items():
        globals()[key] = rxd.Species(r, name=key, initial=value)


k3 = rxd.Parameter(r, name="k3", initial=1.2)
k4 = rxd.Parameter(r, name="k4", initial=0.6)

P2 = rxd.Species(r, name="P2", initial=0.0614368)
T2 = rxd.Species(r, name="T2", initial=0.0145428)
C = rxd.Species(r, name="C", initial=0.207614)

h.finitialize(-65)
h.continuerun(72)



import pytest
import numpy
from testutils import compare_data, tol


@pytest.fixture
def ics_pure_diffusion(neuron_instance):
    """A model using intracellular diffusion in a single section in 3D"""

    h, rxd, data, save_path = neuron_instance
    dend = h.Section(name="dend")
    dend.diam = 2
    dend.nseg = 11
    dend.L = 5
    rxd.set_solve_type(dimension=3)
    diff_constant = 1
    r = rxd.Region(h.allsec(), dx=0.75)
    ca = rxd.Species(
        r, d=diff_constant, initial=lambda node: 1 if 0.4 < node.x < 0.6 else 0
    )
    model = (dend, r, ca)
    yield (neuron_instance, model)


def test_pure_diffusion_3d(ics_pure_diffusion):
    """Test ics_pure_diffusion with fixed step methods"""

    neuron_instance, model = ics_pure_diffusion
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = model
    h.dt *= 50
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_cvode(ics_pure_diffusion):
    """Test ics_pure_diffusion with variable step methods"""

    neuron_instance, model = ics_pure_diffusion
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = model
    h.CVode().active(True)
    vec = h.Vector()
    h.CVode().states(vec)
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_inhom(ics_pure_diffusion):
    """Test ics_pure_diffusion with fixed step methods and inhomogeneous
    diffusion coefficients.
    """
    neuron_instance, model = ics_pure_diffusion
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = model
    h.dt *= 50
    for nd in ca.nodes:
        if nd.x >= 0.5:
            nd.d = 0
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_inhom_cvode(ics_pure_diffusion):
    """Test ics_pure_diffusion with variable step methods and inhomogeneous
    diffusion coefficients.
    """
    neuron_instance, model = ics_pure_diffusion
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = model
    h.CVode().active(True)
    for nd in ca.nodes:
        if nd.x >= 0.5:
            nd.d = 0
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol



#cython: language_level=2
from neuron import h
import numpy
import itertools
import bisect
cimport numpy
from numpy import linalg
cimport cython
from neuron.rxd.rxdException import RxDException
import neuron


"""
The ctng module
"""

cdef extern from "math.h":
    double sqrt(double)
    double fabs(double)

from graphicsPrimitives import Sphere, Cone, Cylinder, SkewCone, Plane, Union, Intersection, SphereCone

cdef tuple seg_line_intersection(double x1, double y1, double x2, double y2, double x3, double y3, double x4, double y4, bint clip):
    # returns None if parallel (so None if 0 or infinitely many intersections)
    # if clip is True, requires intersection on segment; else returns line-line intersection
    cdef double denom = (y4 - y3) * (x2 - x1) - (x4 - x3) * (y2 - y1)
    if denom == 0: return None
    cdef double u = ((x4 - x3) * (y1 - y3) - (y4 - y3) * (x1 - x3)) / denom
    if clip and not (0 <= u <= 1): return None
    return (x1 + u * (x2 - x1), y1 + u * (y2 - y1))

cdef tuple closest_pt(pt, list pts, z2):
    dist = float('inf')
    closest = None
    for p in pts:
        x, y = p
        d = linalg.norm(numpy.array(pt) - numpy.array((x, y, z2)))
        if d < dist:
            dist = d
            closest = p
    return tuple(closest)

cdef tuple extreme_pts(list pts):
    if len(pts) < 2: raise RxDException('extreme points computation failed')
    cdef double max_dist, d
    cdef tuple pt1, pt2, best_p1, best_p2
    max_dist = -1

    for pt1, pt2 in itertools.combinations(pts, 2):
        d = linalg.norm(numpy.array(pt1) - numpy.array(pt2))
        if d > max_dist:
            best_p1 = pt1
            best_p2 = pt2
            max_dist = d
    return best_p1, best_p2

# helper function for maintaing the points-cones database
cdef register(dict pts_cones_db, tuple pt, cone):
    if pt not in pts_cones_db:
        pts_cones_db[pt] = []
    pts_cones_db[pt].append(cone)

# helper function that counts the number of points inside a region
@cython.boundscheck(False)
@cython.wraparound(False)
cdef int count_outside(region, list pts, double err):
    cdef numpy.ndarray[numpy.float_t, ndim=1] pt
    cdef int result = 0
    for pt in pts:
        if region.distance(pt[0], pt[1], pt[2]) > err:
            result += 1
    return result

@cython.boundscheck(False)
@cython.wraparound(False)
cdef tuple get_infinite_cones(double x0, double y0, double z0, double r0, double x1, double y1, double z1, double r1, double x2, double y2, double z2, double r2):
    cdef double deltar, deltanr
    cdef numpy.ndarray[numpy.float_t, ndim=1] axis, naxis

    axis = numpy.array([x2 - x1, y2 - y1, z2 - z1])
    naxis = numpy.array([x1 - x0, y1 - y0, z1 - z0])
    deltar = r2 - r1; deltanr = r1 - r0
    deltar /= linalg.norm(axis); deltanr /= linalg.norm(naxis)
    axis /= linalg.norm(axis); naxis /= linalg.norm(naxis)
    #
    # sphere, clipped to extended cones
    # CTNG:trimsphere
    #
    offset0 = r0
    offset1 = r1
    offset1b = r1
    offset2 = r2
    if r0 - deltanr * offset0 < 0:
        # don't go that far (reduce offset0 so that this comes out 0)
        offset0 = r0 / deltanr
    if r1 + deltanr * offset1 < 0:
        offset1 = -r1 / deltanr
    if r1 - deltar * offset1b < 0:
        offset1b = r1 / deltar
    if r2 + deltar * offset2 < 0:
        offset2 = -r2 / deltar

    c0 = Cone(x0 - naxis[0] * offset0, y0 - naxis[1] * offset0, z0 - naxis[2] * offset0, r0 - deltanr * offset0, x1 + naxis[0] * offset1, y1 + naxis[1] * offset1, z1 + naxis[2] * offset1, r1 + deltanr * offset1)
    c1 = Cone(x1 - axis[0] * offset1b, y1 - axis[1] * offset1b, z1 - axis[2] * offset1b, r1 - deltar * offset1b, x2 + axis[0] * offset2, y2 + axis[1] * offset2, z2 + axis[2] * offset2, r2 + deltar * offset2)
    return c0, c1


@cython.boundscheck(False)
@cython.wraparound(False)
cdef list join_outside(double x0, double y0, double z0, double r0, double x1, double y1, double z1, double r1, double x2, double y2, double z2, double r2, double dx):
    cdef numpy.ndarray[numpy.float_t, ndim=1] pt1, radial_vec, nradial_vec, axis, naxis

    c0, c1 = get_infinite_cones(x0, y0, z0, r0, x1, y1, z1, r1, x2, y2, z2, r2)

    axis = numpy.array([x2 - x1, y2 - y1, z2 - z1])
    naxis = numpy.array([x1 - x0, y1 - y0, z1 - z0])
    axis /= linalg.norm(axis); naxis /= linalg.norm(naxis)
    #
    # sphere, clipped to extended cones
    # CTNG:trimsphere
    #
    sp = Sphere(x1, y1, z1, r1)
    sp.set_clip([Intersection([c0, c1])])

    result = [sp]
    # check to see if the clipped sphere covers the ends of the cones
    # if not, do something else :)

    # locate key vectors
    plane_normal = numpy.cross(axis, naxis)
    radial_vec = numpy.cross(plane_normal, axis)
    nradial_vec = numpy.cross(plane_normal, naxis)
    # normalize all of these
    radial_vec /= linalg.norm(radial_vec)
    nradial_vec /= linalg.norm(nradial_vec)

    # count the corners that are inside a sphere clipped to the cones
    pt1 = numpy.array([x1, y1, z1])
    cdef int left_corner_count = 2 - count_outside(sp, [pt1 + r1 * nradial_vec, pt1 - r1 * nradial_vec], dx * 0.5)
    cdef int corner_count = 2 - count_outside(sp, [pt1 + r1 * radial_vec, pt1 - r1 * radial_vec], dx * 0.5)
    #print 'for join (%g, %g, %g; %g) - (%g, %g, %g; %g) - (%g, %g, %g; %g):' % (x0, y0, z0, r0, x1, y1, z1, r1, x2, y2, z2, r2)
    #print '   left_corner_count = %g; corner_count = %g' % (left_corner_count, corner_count)
    if left_corner_count == corner_count == 2:
        sp.set_clip([Intersection([c0, c1,
            Plane(x1, y1, z1, axis[0], axis[1], axis[2]),
            Plane(x1, y1, z1, -naxis[0], -naxis[1], -naxis[2])])])
    elif left_corner_count < 2 and corner_count == 2:
        # clipping to c1 is too harsh, but c0 clip is fine
        sp.set_clip([Intersection([c0,
            Plane(x1, y1, z1, axis[0], axis[1], axis[2]),
            Plane(x1, y1, z1, -naxis[0], -naxis[1], -naxis[2])])])
    elif left_corner_count == 2 and corner_count < 2:
        # clipping to c0 is too harsh, but c1 clip is fine
        sp.set_clip([Intersection([c1,
            Plane(x1, y1, z1, axis[0], axis[1], axis[2]),
            Plane(x1, y1, z1, -naxis[0], -naxis[1], -naxis[2])])])
    elif left_corner_count < 2 and corner_count < 2:
        # both clips are too harsh; fall back to just using a sphere
        sp.set_clip([Intersection([
            Plane(x1, y1, z1, axis[0], axis[1], axis[2]),
            Plane(x1, y1, z1, -naxis[0], -naxis[1], -naxis[2])])])
    else:
        raise RxDException('unexpected corner_counts?')

    return result

@cython.wraparound(False)
@cython.boundscheck(False)
def soma_objects(x, y, z, sec, double x0, double y0, double z0, int n_soma_step):
    cdef double diam1, diam2, somax, somay, somaz
    cdef list objects = []
    cdef list f_pts
    cdef dict seg_dict = {}
    
    somax, somay, somaz = x0, y0, z0

    xshifted = [xx - x0 for xx in x]
    yshifted = [yy - y0 for yy in y]
    # this is a hack to pretend everything is on the same z level
    zshifted = [0] * len(x)

    # locate the major and minor axis, adapted from import3d_gui.hoc
    m = h.Matrix(3, 3)
    for i, p in enumerate([xshifted, yshifted, zshifted]):
        for j, q in enumerate([xshifted, yshifted, zshifted]):
            if j < i: continue
            v = numpy.dot(p, q)
            m.setval(i, j, v)
            m.setval(j, i, v)
    # CTNG:majoraxis
    tobj = m.symmeig(m)
    # major axis is the one with largest eigenvalue
    major = m.getcol(tobj.max_ind())
    # minor is normal and in xy plane
    minor = m.getcol(3 - tobj.min_ind() - tobj.max_ind())
    #minor.x[2] = 0
    minor.div(minor.mag())

    x1 = x0; y1 = y0
    x2 = x1 + major.x[0]; y2 = y1 + major.x[1]

    xs_loop = x + [x[0]]
    ys_loop = y + [y[0]]

    # locate the extrema of the major axis CTNG:somaextrema
    # this is defined by the furthest points on it that lie on the minor axis
    pts = []
    pts_sources = {}
    for x3, y3 in zip(x, y):
        x4, y4 = x3 + minor.x[0], y3 + minor.x[1]
        pt = seg_line_intersection(x1, y1, x2, y2, x3, y3, x4, y4, False)
        if pt is not None:
            pts.append(pt)
            if pt not in pts_sources:
                pts_sources[pt] = []
            pts_sources[pt].append((x3, y3))

    major_p1, major_p2 = extreme_pts(pts)

    extreme1 = pts_sources[major_p1]
    extreme2 = pts_sources[major_p2]

    major_p1, major_p2 = numpy.array(major_p1), numpy.array(major_p2)
    del pts_sources

    if len(extreme1) != 1 or len(extreme2) != 1:
        raise RxDException('multiple most extreme points')
    extreme1 = extreme1[0]
    extreme2 = extreme2[0]
    major_length = linalg.norm(major_p1 - major_p2)
    delta_x, delta_y = major_p2 - major_p1
    delta_x /= n_soma_step
    delta_y /= n_soma_step

    f_pts = [extreme1]
    f_diams = [0]

    # CTNG:slicesoma
    for i in xrange(1, n_soma_step):
        x0, y0 = major_p1[0] + i * delta_x, major_p1[1] + i * delta_y
        # slice in dir of minor axis
        x1, y1 = x0 + minor.x[0], y0 + minor.x[1]
        pts = []
        for i in xrange(len(x)):
            pt = seg_line_intersection(xs_loop[i], ys_loop[i], xs_loop[i + 1], ys_loop[i + 1], x0, y0, x1, y1, True)
            if pt is not None: pts.append(pt)
        p1, p2 = extreme_pts(pts)
        p1, p2 = numpy.array(p1), numpy.array(p2)
        cx, cy = (p1 + p2) / 2.
        f_pts.append((cx, cy))
        f_diams.append(linalg.norm(p1 - p2))

    f_pts.append(extreme2)
    f_diams.append(0)

    segment_locs = [];        # location along soma for assigning segments
    path_length = 0;
    margin = 1/(sec.nseg*2)   # half a segment, normalized length
    for i in xrange(len(f_pts) - 1):
        pt1x, pt1y = f_pts[i]
        pt2x, pt2y = f_pts[i + 1]
        diam1 = f_diams[i]
        diam2 = f_diams[i + 1]
        thickness = numpy.sqrt((pt2x-pt1x)**2 + (pt2y-pt1y)**2)
        segment_locs.append(path_length+(thickness/2))
        path_length += thickness

        objects.append(SkewCone(pt1x, pt1y, z0, diam1 * 0.5, pt1x + delta_x, pt1y + delta_y, z0, diam2 * 0.5, pt2x, pt2y, z0))
    for j, s in enumerate(objects):
        loc = segment_locs[j]/path_length       # normalized along length
        for seg in sec:
            if numpy.abs(seg.x - loc) <= margin:
                seg_dict[s] = seg
                break;
    return seg_dict, f_pts
@cython.wraparound(False)
@cython.boundscheck(False)
def constructive_neuronal_geometry(source, int n_soma_step, double dx, nouniform=False, relevant_pts=None):
    cdef list objects = []
    cdef dict cone_segment_dict = {}
    cdef dict soma_segment_dict = {}
    cdef list join_groups = []
    cdef dict obj_pts_dict = {}         # for returning corner points of join objects
    cdef list obj_sections = []
    cdef list cone_sections = []
    cdef dict soma_secs = {}
    cdef dict potential_soma_cones = {}
    cdef int i, j, k
    cdef double x0, y0, z0, x1, y1, z1, x2, y2, z2, x3, y3, x4, y4, r0, r1, r2
    cdef double delta_x, delta_y, major_length, diam1, diam2
    cdef list pts, f_pts, f_diams, branches, parent_sec_name
    cdef dict pts_sources
    cdef tuple pt
    #cdef numpy.ndarray[numpy.float_t, ndim=1] x, y, z, xs_loop, ys_loop
    cdef int no_parent_count = 0

    source_is_import3d = False
    branches = []
    f_pts = []
    parent_sec_name = []
    num_contours = None
    # TODO: come up with a better way of checking type
    if hasattr(source, 'sections'):
        source_is_import3d = True
        cell = source
        # probably an Import3D type
        num_contours = sum(sec.iscontour_ for sec in cell.sections)
        if num_contours > 1:
            raise RxDException('more than one contour is not currently supported')
        if num_contours <= 1:
            # CTNG:soma
            branches = []
            parent_sec_name = []
            for sec in cell.sections:
                if sec.iscontour_:
                    x, y, z = [sec.raw.getrow(i).to_python() for i in xrange(3)]

                    # compute the center of the contour based on uniformly spaced points around the perimeter
                    center_vec = sec.contourcenter(sec.raw.getrow(0), sec.raw.getrow(1), sec.raw.getrow(2))
                    x0, y0, z0 = [center_vec.x[i] for i in xrange(3)]
                    new_objects, f_pts = soma_objects(x, y, z, sec, x0, y0, z0, n_soma_step)
                    soma_secs[sec] = (f_pts, z0)
                    soma_segment_dict.update(new_objects)
                else:
                    if sec.parentsec is not None:
                        parent_sec_name.append(sec.parentsec.hname())
                    else:
                        parent_sec_name.append(no_parent_count)
                        no_parent_count += 1
                    branches.append(sec)

    else:
        h.define_shape()
        branches = []
        for sec in source:
            # TODO: make this more general (support for 3D contour outline)
            if sec.hoc_internal_name() in neuron._sec_db:
                is_stack, x, y, z, x0, y0, z0, pts3d = neuron._sec_db[sec.hoc_internal_name()]
                if not is_stack:
                    shift = []
                    for i, (orig_x, orig_y, orig_z) in zip([0, sec.n3d()-1], pts3d):
                        shift.append((sec.x3d(i) - orig_x,
                                      sec.y3d(i) - orig_y,
                                      sec.z3d(i) - orig_z))
                    if sum([abs(s0 - s1) for s0,s1 in zip(*shift)]) > dx/10.0:
                        raise RxDException("soma rotation unsupported for voxelized somas") 
                    sx, sy, sz = shift[0]
                    x = (x + sx).to_python()
                    y = (y + sy).to_python()
                    z = (z + sz).to_python()
                    new_objects, f_pts = soma_objects(x, y, z, sec, x0 + sx, y0 + sy, z0 + sz, n_soma_step)
                    soma_secs[sec] = (f_pts, z0 + sz)
                    soma_segment_dict.update(new_objects)
                else:
                    import warnings
                    warnings.warn('soma stack ignored; using centroid instead')
                    branches.append(sec)
                    parent_sec_name.append(None)
            else:
                branches.append(sec)
                if sec.trueparentseg():
                    parent_sec_name.append(sec.trueparentseg().sec)
                else:
                    parent_sec_name.append(None)


    #####################################################################
    #
    # add the branches
    #
    #####################################################################
    cdef dict diam_corrections = {None: None}
    cdef dict clip_copying = {}

    while diam_corrections:
        all_cones = []
        pts_cones_db = {}
        diam_db = {}
        
        for (k, branch), psec in zip(enumerate(branches), parent_sec_name):
            if source_is_import3d:
                x, y, z = [numpy.array(branch.raw.getrow(i).to_python()) for i in range(3)]
                d = branch.d.to_python()
            else:
                #################### modified from geometry.py ##################
                # make sure cones split across electrical segments
                if relevant_pts:
                    rng = relevant_pts[k]
                else:
                    rng = range(branch.n3d())

                arc3d = ([branch.arc3d(i) for i in rng])
                diam3d = ([branch.diam3d(i) for i in rng])
                x3d = ([branch.x3d(i) for i in rng])
                y3d = ([branch.y3d(i) for i in rng])
                z3d = ([branch.z3d(i) for i in rng])

                x = numpy.array([])
                y = numpy.array([])
                z = numpy.array([])
                d = numpy.array([])

                dx = branch.L / branch.nseg
                for iseg, seg in enumerate(branch):
                    # get a list of all pts in the segment, including end points
                    lo = iseg * dx
                    hi = (iseg + 1) * dx

                    pts = [lo] + [arc for arc in arc3d if lo < arc < hi] + [hi]

                    diams = numpy.interp(pts, arc3d, diam3d)
                    xcoords = numpy.interp(pts, arc3d, x3d)
                    ycoords = numpy.interp(pts, arc3d, y3d)
                    zcoords = numpy.interp(pts, arc3d, z3d)

                    end = len(pts)-1
                    if iseg == branch.nseg-1:
                        end = len(pts)

                    x = numpy.append(x, xcoords[0:end])
                    y = numpy.append(y, ycoords[0:end])
                    z = numpy.append(z, zcoords[0:end])
                    d = numpy.append(d, diams[0:end])

                    for i in range(len(pts)-1):
                        #find the cone and assign it to that segment
                        conecoords = (xcoords[i],ycoords[i],zcoords[i],xcoords[i+1],ycoords[i+1],zcoords[i+1])
                        cone_segment_dict[conecoords] = seg
                ##########################################################


            # make sure that all the ones that connect to the soma do in fact connect
            # do this by connecting to local center axis
            # CTNG:connectdends

            if psec in soma_secs:
                f_pts, somaz = soma_secs[psec]
                pt = (x[1], y[1], z[1])
                cp = closest_pt(pt, f_pts, somaz)
                # NEURON includes the wire point at the center; we want to connect
                # to the closest place on the soma's axis instead with full diameter
                # x, y, z, d = [cp[0]] + [X for X in x[1 :]], [cp[1]] + [Y for Y in y[1:]], [somaz] + [Z for Z in z[1:]], [d[1]] + [D for D in d[1 :]]
                x[0], y[0] = cp
                z[0] = somaz 
                d[0] = d[1]
                if branch not in potential_soma_cones:
                    potential_soma_cones[branch] = []

                '''# cap this with a sphere for smooth joins
                sphere_cap = Sphere(x[0], y[0], z[0], d[0] * 0.5)
                # make sure sphere doesn't stick out of wrong side of cylinder
                sphere_cap.set_clip([Plane(x[1], y[1], z[1], x[1] - x[0], y[1] - y[0], z[1] - z[0])])
                objects.append(sphere_cap)
                with cython.wraparound(True):
                    soma_objects.append(objects[-1])'''


            for i in range(len(x) - 1):
                d0, d1 = d[i : i + 2]
                if (x[i] != x[i + 1] or y[i] != y[i + 1] or z[i] != z[i + 1]):
                    # short section check
                    #if linalg.norm((x[i + 1] - x[i], y[i + 1] - y[i], z[i + 1] - z[i])) < (d1 + d0) * 0.5:
                    #    short_segs += 1
                    axisx, axisy, axisz, deltad = x[i + 1] - x[i], y[i + 1] - y[i], z[i + 1] - z[i], d1 - d0
                    axislength = (axisx ** 2 + axisy ** 2 + axisz ** 2) ** 0.5
                    axisx /= axislength; axisy /= axislength; axisz /= axislength; deltad /= axislength
                    x0, y0, z0 = x[i], y[i], z[i]
                    x1, y1, z1 = x[i + 1], y[i + 1], z[i + 1]
                    if (x0, y0, z0) in diam_corrections:
                        d0 = diam_corrections[(x0, y0, z0)]
                    if (x1, y1, z1) in diam_corrections:
                        d1 = diam_corrections[(x1, y1, z1)]

                    if d0 != d1:
                        all_cones.append(Cone(x0, y0, z0, d0 * 0.5, x1, y1, z1, d1 * 0.5))
                    else:
                        all_cones.append(Cylinder(x0, y0, z0, x1, y1, z1, d1 * 0.5))
                    with cython.wraparound(True):
                        register(pts_cones_db, (x0, y0, z0), all_cones[-1])
                        register(pts_cones_db, (x1, y1, z1), all_cones[-1])
                        # if the cone is added to connect the dendrite to the soma
                        # and it does not have corresponding segment
                        # keep track of them here and assign them to a soma
                        # segment later.
                        if branch in potential_soma_cones:
                            potential_soma_cones[branch].append(all_cones[-1])
                    register(diam_db, (x0, y0, z0), d0)
                    register(diam_db, (x1, y1, z1), d1)

        diam_corrections = {}
        if not nouniform:
            # at join, should always be the size of the biggest branch
            # this is different behavior than NEURON, which continues the size of the
            # first point away from the join to the join
            for pt in diam_db:
                vals = diam_db[pt]
                if max(vals) != min(vals):
                    diam_corrections[pt] = max(vals)



    cdef dict cone_clip_db = {cone: [] for cone in all_cones}
    cdef bint sharp_turn
#    cdef dict join_counts = {'2m': 0, '2s': 0, '3m': 0, '3s': 0, '4m': 0, '4s': 0, '0m': 0, '0s': 0, '1m': 0, '1s': 0}
    join_items_needing_clipped = []
    for cone in all_cones:
        joingroup = []
        x1, y1, z1, r1 = cone._x0, cone._y0, cone._z0, cone._r0
        x2, y2, z2, r2 = cone._x1, cone._y1, cone._z1, cone._r1
        pt1 = numpy.array([x1, y1, z1])
        pt2 = numpy.array([x2, y2, z2])
        axis = (pt2 - pt1) / linalg.norm(pt2 - pt1)
        left_neighbors = list(pts_cones_db[(x1, y1, z1)])

        right_neighbors = list(pts_cones_db[(x2, y2, z2)])
        left_neighbors.remove(cone)
        right_neighbors.remove(cone)
        if not left_neighbors:
            left_neighbors = [None]
        else:
            joingroup.append(cone)
            for neighbor in left_neighbors:
                joingroup.append(neighbor)
        if not right_neighbors: right_neighbors = [None]
        for neighbor_left, neighbor_right in itertools.product(left_neighbors, right_neighbors):
            clips = []
            # if any join needs to be subject to clips, it goes here
            join_item = None
            # process the join on the "left" (end 1)
            if neighbor_left is not None:
                # any joins are created on the left pass; the right pass will only do clippings
                x0, y0, z0, r0 = neighbor_left._x0, neighbor_left._y0, neighbor_left._z0, neighbor_left._r0
                if x0 == x1 and y0 == y1 and z0 == z1:
                    x0, y0, z0, r0 = neighbor_left._x1, neighbor_left._y1, neighbor_left._z1, neighbor_left._r1
                pt0 = numpy.array([x0, y0, z0])
                naxis = (pt1 - pt0) / linalg.norm(pt1 - pt0)
                # no need to clip if the cones are perfectly aligned

                if all(axis == naxis):
                    #parallel
                    sp = Sphere(x1, y1, z1, r1)
                    c0, c1 = get_infinite_cones(x0, y0, z0, r0, x1, y1, z1, r1, x2, y2, z2, r2)

                    sp.set_clip([c0, c1, Plane(x0, y0, z0, -naxis[0], -naxis[1], -naxis[2]), Plane(x2, y2, z2, axis[0], axis[1], axis[2])])
                    objects.append(sp)
                    with cython.wraparound(True):
                        joingroup.append(objects[-1])
                elif r0 == r1 == r2:
                    # simplest join: two non-parallel cylinders (no need for all that nastiness below)
                    sp = Sphere(x1, y1, z1, r1)
                    c0, c1 = get_infinite_cones(x0, y0, z0, r0, x1, y1, z1, r1, x2, y2, z2, r2)
                    sp.set_clip([Intersection([c0, c1, Plane(x0, y0, z0, -naxis[0], -naxis[1], -naxis[2]), Plane(x2, y2, z2, axis[0], axis[1], axis[2])])])
                    objects.append(sp)
                    with cython.wraparound(True):
                        joingroup.append(objects[-1])
                else:
                    # is the turn sharp or not
                    # CTNG:joinangle
                    sharp_turn = numpy.dot(axis, naxis) < 0
                    # locate key vectors
                    plane_normal = numpy.cross(axis, naxis)
                    radial_vec = numpy.cross(plane_normal, axis)
                    nradial_vec = numpy.cross(plane_normal, naxis)
                    # normalize all of these
                    radial_vec /= linalg.norm(radial_vec)
                    nradial_vec /= linalg.norm(nradial_vec)

                    # count the corners that are inside the other cone (for both ways)
                    # CTNG:outsidecorners
                    corner_pts = [pt1 + r1 * radial_vec, pt1 - r1 * radial_vec, pt1 + r1 * nradial_vec, pt1 - r1 * nradial_vec]
                    my_corner_count = count_outside(neighbor_left, [corner_pts[0], corner_pts[1]], 0)
                    corner_count = my_corner_count + count_outside(cone, [corner_pts[2], corner_pts[3]], 0)


                    # if corner_count == 0, then probably all nan's from size 0 meeting size 0; ignore
                    # if is 1, probably parallel; no joins
    #                if corner_count not in (1, 2, 3, 4):
    #                    print 'corner_count: ', corner_count, [pt1 + r1 * radial_vec, pt1 - r1 * radial_vec] + [pt1 + r1 * nradial_vec, pt1 - r1 * nradial_vec]
                    if corner_count == 2:
                        # CTNG:2outside
                        # add clipped sphere; same rule if sharp or mild turn
                        objects += join_outside(x0, y0, z0, r0, x1, y1, z1, r1, x2, y2, z2, r2, dx)
                        with cython.wraparound(True):
                            joingroup.append(objects[-1])
                            obj_pts_dict[objects[-1]] = corner_pts
                    elif corner_count == 3:
                        sp = Sphere(x1, y1, z1, r1)
                        if sharp_turn:
                            # CTNG:3outobtuse
                            if my_corner_count == 1:
                                sp.set_clip([Plane(x1, y1, z1, -naxis[0], -naxis[1], -naxis[2])])
                            else:
                                sp.set_clip([Plane(x1, y1, z1, axis[0], axis[1], axis[2])])
                            objects.append(sp)
                            with cython.wraparound(True):
                                joingroup.append(objects[-1])
                                obj_pts_dict[objects[-1]] = corner_pts
                        else:
                            # CTNG:3outacute
                            objects += join_outside(x0, y0, z0, r0, x1, y1, z1, r1, x2, y2, z2, r2, dx)
                            with cython.wraparound(True):
                                joingroup.append(objects[-1])
                                obj_pts_dict[objects[-1]] = corner_pts
                                if my_corner_count == 1:
                                    objects.append(sp)
                                    objects[-1].set_clip([Plane(x0, y0, z0, -axis[0], -axis[1], -axis[2]), Plane(x2, y2, z2, naxis[0], naxis[1], naxis[2])])
                                    joingroup.append(objects[-1])
                                    obj_pts_dict[objects[-1]] = corner_pts
                                else:
                                    objects.append(sp)
                                    objects[-1].set_clip([Plane(x0, y0, z0, -axis[0], -axis[1], -axis[2]), Plane(x2, y2, z2, naxis[0], naxis[1], naxis[2])])
                                    joingroup.append(objects[-1])
                                    obj_pts_dict[objects[-1]] = corner_pts



                    elif corner_count == 4:
                        sp = Sphere(x1, y1, z1, r1)
                        if sharp_turn:
                            # CTNG:4outobtuse
                            # join with the portions of a sphere that are outside at least one of the planes
                            sp.set_clip([Union([
                                Plane(x1, y1, z1, axis[0], axis[1], axis[2]),
                                Plane(x1, y1, z1, -naxis[0], -naxis[1], -naxis[2])])])
                            objects.append(sp)
                            with cython.wraparound(True):
                                joingroup.append(objects[-1])
                                obj_pts_dict[objects[-1]] = corner_pts
                        else:
                            # CTNG:4outacute (+ 1 more)
                            # join with the portions of a sphere that are outside both planes
                            objects += join_outside(x0, y0, z0, r0, x1, y1, z1, r1, x2, y2, z2, r2, dx)
                            with cython.wraparound(True):
                                joingroup.append(objects[-1])
                                obj_pts_dict[objects[-1]] = corner_pts
                            # AND clip the cone to not extend pass the union of the neighbor's plane and the neighbor
                            if r0 == r1:
                                neighbor_copy = Cylinder(x0, y0, z0, x1, y1, z1, r0)
                            else:
                                neighbor_copy = Cone(x0, y0, z0, r0, x1, y1, z1, r1)
                            clips.append(Union([
                                Plane(x1, y1, z1, -naxis[0], -naxis[1], -naxis[2]),
                                Plane(x0, y0, z0, axis[0], axis[1], axis[2])]))

#                        join_type = '%d%s' % (corner_count, 's' if sharp_turn else 'm')
#                        join_counts[join_type] += 1





            if neighbor_right is not None:
                # any joins are created on the left pass; the right pass will only do clippings
                x3, y3, z3, r3 = neighbor_right._x0, neighbor_right._y0, neighbor_right._z0, neighbor_right._r0
                if x2 == x3 and y2 == y3 and z2 == z3:
                    x3, y3, z3, r3 = neighbor_right._x1, neighbor_right._y1, neighbor_right._z1, neighbor_right._r1
                pt3 = numpy.array([x3, y3, z3])
                naxis = (pt3 - pt2) / linalg.norm(pt3 - pt2)

                # no need to clip if the cones are perfectly aligned
                if any(axis != naxis):
                    # locate key vectors
                    plane_normal = numpy.cross(axis, naxis)
                    radial_vec = numpy.cross(plane_normal, axis)
                    radial_vec_norm = linalg.norm(radial_vec)

                    # we check again because sometimes there are roundoff errors that this catches
                    if radial_vec_norm:

                        # is the turn sharp or not
                        sharp_turn = numpy.dot(axis, naxis) < 0

                        nradial_vec = numpy.cross(plane_normal, naxis)
                        # normalize all of these
                        radial_vec /= radial_vec_norm
                        nradial_vec /= linalg.norm(nradial_vec)
                        # count the corners that are inside the other cone (for both ways)
                        my_corner_count = count_outside(neighbor_right, [pt2 + r2 * radial_vec, pt2 - r2 * radial_vec], 0)
                        corner_count = my_corner_count + count_outside(cone, [pt2 + r2 * nradial_vec, pt2 - r2 * nradial_vec], 0)
                        if corner_count == 2:
                            # no clipping; already joined
                            pass
                        elif corner_count == 3:
                            pass

                        elif corner_count == 4:
                            # CTNG:4outacute (+ 1 more)
                            # already joined; just clip (only in mild turn case)
                            if not sharp_turn:
                                if r2 == r3:
                                    neighbor_copy = Cylinder(x2, y2, z2, x3, y3, z3, r3)
                                else:
                                    neighbor_copy = Cone(x2, y2, z2, r2, x3, y3, z3, r3)
                                #print 'cc=4: (%g, %g, %g; %g) (%g, %g, %g; %g) (%g, %g, %g; %g) ' % (x1, y1, z1, r1, x2, y2, z2, r2, x3, y3, z3, r3)
                                clips.append(Union([
                                    Plane(x2, y2, z2, naxis[0], naxis[1], naxis[2]),
                                    neighbor_copy]))



            if clips:
                int_clip = Intersection(clips)
                cone_clip_db[cone].append(int_clip)

        if joingroup:
            join_groups.append(joingroup)

    for cone in all_cones:
        clip = cone_clip_db[cone]
        if clip:
            cone.set_clip([Union(clip)])

    for sec,cones in potential_soma_cones.items():
        for cone in cones:
            if (cone._x0, cone._y0, cone._z0, cone._x1, cone._y1, cone._z1) not in cone_segment_dict:
                cone_segment_dict[(cone._x0, cone._y0, cone._z0, cone._x1, cone._y1, cone._z1)] = sec.trueparentseg()

    #####################################################################
    #
    # add the clipped objects to the list
    #
    #####################################################################

    #objects += all_cones

    return [objects, all_cones, cone_segment_dict, join_groups, obj_pts_dict, soma_segment_dict]



from testutils import compare_data, tol


def test_multicompartment_reactions(neuron_instance):
    """A tests of mulicompartment reactions using an intracellular Ca model.

    Test based on example copied from the RxD tutorial;
    http://www.neuron.yale.edu/neuron/static/docs/rxd/index.html
    Where 1D intracellular space is divided into cytsol and endoplasmic
    reticulum. Calcium is transported between the regions by multicompartment
    reactions for a leak, SERCA pump and IP3 receptor.
    """

    h, rxd, data, save_path = neuron_instance
    sec = h.Section()
    sec.L = 100
    sec.diam = 1
    sec.nseg = 100

    h.CVode().active(1)
    h.CVode().atol(1e-4)

    caDiff = 0.016
    ip3Diff = 0.283
    cac_init = 1.0e-4
    ip3_init = 0.1
    gip3r = 12040
    gserca = 0.3913
    gleak = 6.020
    kserca = 0.1
    kip3 = 0.15
    kact = 0.4
    ip3rtau = 2000.0

    # These parameters where missing in the tutorial so arbitrary values were chosen
    # any resemblance to experimental values is purely coincidental.
    fc = 0.7
    fe = 0.3
    caCYT_init = 0.1

    cyt = rxd.Region(
        h.allsec(),
        name="cyt",
        nrn_region="i",
        geometry=rxd.FractionalVolume(fc, surface_fraction=1),
    )

    er = rxd.Region(h.allsec(), name="er", geometry=rxd.FractionalVolume(fe / 2.0))
    cyt_er_membrane = rxd.Region(
        h.allsec(), name="mem", geometry=rxd.ScalableBorder(1, on_cell_surface=False)
    )
    ca = rxd.Species([cyt, er], d=caDiff, name="ca", charge=2, initial=caCYT_init)

    ip3 = rxd.Species(cyt, d=ip3Diff, name="ip3", initial=ip3_init)
    ip3r_gate_state = rxd.Species(cyt_er_membrane, name="gate", initial=0.8)
    minf = ip3[cyt] * 1000.0 * ca[cyt] / (ip3[cyt] + kip3) / (1000.0 * ca[cyt] + kact)
    k = gip3r * (minf * ip3r_gate_state[cyt_er_membrane]) ** 3

    ip3r = rxd.MultiCompartmentReaction(ca[er], ca[cyt], k, k, membrane=cyt_er_membrane)

    serca = rxd.MultiCompartmentReaction(
        ca[cyt],
        ca[er],
        gserca / ((kserca / (1000.0 * ca[cyt])) ** 2 + 1),
        membrane=cyt_er_membrane,
        custom_dynamics=True,
    )
    leak = rxd.MultiCompartmentReaction(
        ca[er], ca[cyt], gleak, gleak, membrane=cyt_er_membrane
    )
    # test the SpeciesOnRegion remains in scope
    ip3rg = rxd.Rate(
        ip3r_gate_state[cyt_er_membrane],
        (1.0 / (1 + 1000.0 * ca[cyt] / (0.3)) - ip3r_gate_state[cyt_er_membrane])
        / ip3rtau,
    )

    h.finitialize(-65)

    cae_init = (0.0017 - cac_init * fc) / fe
    ca[er].concentration = cae_init

    for node in ip3.nodes:
        if node.x < 0.2:
            node.concentration = 2

    h.CVode().re_init()
    h.continuerun(1000)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol



import pytest

from testutils import compare_data, tol


@pytest.fixture
def ecs_include_flux(neuron_instance):
    """A model using node.include_flux for the extracellular space.

    node.include_flux can take either a; integer, Python function or NEURON
    pointer. All three are tested here for extracellular rxd.
    """

    h, rxd, data, save_path = neuron_instance
    sec = h.Section(name="dend")
    # the extracellular space
    ecs = rxd.Extracellular(
        -55, -55, -55, 55, 55, 55, dx=33, volume_fraction=0.2, tortuosity=1.6
    )

    # Who?
    x = rxd.Species([ecs], name="x", d=0.0, charge=1, initial=0)

    def callbackfun():
        return 1000

    x[ecs].node_by_location(-40, 0, 0).include_flux(1000)

    x[ecs].node_by_location(0, 0, 0).include_flux(callbackfun)

    x[ecs].node_by_location(40, 0, 0).include_flux(sec(0.5)._ref_v)

    yield (neuron_instance, (sec, ecs, x, callbackfun))


def test_ecs_include_flux(ecs_include_flux):
    """Test ecs_include_flux with fixed step methods"""

    neuron_instance, model = ecs_include_flux
    h, rxd, data, save_path = neuron_instance
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_include_flux_cvode(ecs_include_flux):
    """Test ecs_include_flux with variable step methods"""

    neuron_instance, model = ecs_include_flux
    h, rxd, data, save_path = neuron_instance
    h.CVode().active(True)
    h.finitialize(1000)
    h.continuerun(10)

    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol



import weakref
from neuron import h, nrn
from . import node, rxdsection, nodelist
import numpy
from .rxdException import RxDException
import warnings

# all concentration ptrs and indices
_all_cptrs = []
_all_cindices = []
_c_ptr_vector = None
_c_ptr_vector_storage_nrn = None
_c_ptr_vector_storage = None
_last_c_ptr_length = None


def _donothing():
    pass


class _SectionLookup:
    class Lookup:
        def __init__(self):
            self.rxd_sec_list = {}
            self.nrn_sec_list = h.SectionList()

    _instance = None

    def __init__(self):
        if _SectionLookup._instance is None:
            _SectionLookup._instance = _SectionLookup.Lookup()

    def __getitem__(self, key):
        if (
            key in _SectionLookup._instance.nrn_sec_list
            and key.hoc_internal_name() in _SectionLookup._instance.rxd_sec_list
        ):
            return _SectionLookup._instance.rxd_sec_list[key.hoc_internal_name()]
        return []

    def __setitem__(self, key, value):
        _SectionLookup._instance.nrn_sec_list.append(key)
        _SectionLookup._instance.rxd_sec_list[key.hoc_internal_name()] = value

    def __delitem__(self, key):
        _SectionLookup._instance.nrn_sec_list.remove(key)
        if key.hoc_internal_name() in _SectionLookup._instance.rxd_sec_list:
            del _SectionLookup._instance.rxd_sec_list[key.hoc_internal_name()]

    def __iter__(self):
        return iter(_SectionLookup._instance.nrn_sec_list)

    def values(self):
        return _SectionLookup._instance.rxd_sec_list.values()

    def items(self):
        res = {}
        for sec in _SectionLookup._instance.nrn_sec_list:
            res[sec] = _SectionLookup._instance.rxd_sec_list[sec.hoc_internal_name()]
        return res.items()

    def remove(self, rxdsec):
        for key, val in _SectionLookup._instance.rxd_sec_list.items():
            if val == rxdsec:
                del _SectionLookup._instance.rxd_sec_list[key]


def add_values(mat, i, js, vals):
    mat_i = mat[i]
    for (j, val) in zip(js, vals):
        if val == 0:
            continue
        if j in mat_i:
            mat_i[j] += val
            # if mat_i[j] == 0:
            #    del mat_i[j]
        else:
            mat_i[j] = val


def _parent(sec):
    """Return the parent of seg or None if sec is a root"""
    if sec:
        seg = sec.trueparentseg()
        if seg:
            return seg
        else:
            seg = sec.parentseg()
        if seg:
            par = seg.sec
            parseg = par.parentseg()
            if parseg and seg.x == par.orientation:
                # connection point belongs to temp's ancestor
                return _parent(seg.sec)
            else:
                return seg
        else:
            return None
    return None


def _purge_cptrs():
    """purges all cptr information"""

    global _all_cptrs, _all_cindices, _c_ptr_vector, _last_c_ptr_length
    global _c_ptr_vector_storage, _c_ptr_vector_storage_nrn
    _all_cptrs = []
    _all_cindices = []
    _c_ptr_vector = None
    _c_ptr_vector_storage_nrn = None
    _c_ptr_vector_storage = None
    _last_c_ptr_length = None


def _transfer_to_legacy():
    global _c_ptr_vector, _c_ptr_vector_storage, _c_ptr_vector_storage_nrn
    global _last_c_ptr_length

    size = len(_all_cptrs)
    if _last_c_ptr_length != size:
        if size:
            _c_ptr_vector = h.PtrVector(size)
            _c_ptr_vector.ptr_update_callback(_donothing)
            for i, ptr in enumerate(_all_cptrs):
                _c_ptr_vector.pset(i, ptr)
            _c_ptr_vector_storage_nrn = h.Vector(size)
            _c_ptr_vector_storage = _c_ptr_vector_storage_nrn.as_numpy()
        else:
            _c_ptr_vector = None
        _last_c_ptr_length = size
    if size:
        _c_ptr_vector_storage[:] = node._get_states()[_all_cindices]
        _c_ptr_vector.scatter(_c_ptr_vector_storage_nrn)


def replace(rmsec, offset, nseg):
    """Replace the section (rmsec) in node data lists and update the offsets"""
    # remove entries from node global variables and update the states
    node._replace(rmsec._offset, rmsec._nseg, offset, nseg)

    # correct the offsets for all section1ds
    old_offset = rmsec._offset
    dur = rmsec._nseg + 1
    rmsec._offset = offset
    rmsec._nseg = nseg

    for secs in _SectionLookup().values():
        for sec in secs:
            if sec and sec._offset > old_offset:
                sec._offset -= dur


class Section1D(rxdsection.RxDSection):
    def __init__(self, species, sec, diff, r):
        self._species = weakref.ref(species)
        self._diff = diff
        self._secref = h.SectionList([sec])
        self._concentration_ptrs = None
        self._offset = node._allocate(sec.nseg + 1)
        self._nseg = sec.nseg
        self._region = r
        # NOTE: you must do _init_diffusion_rates after assigning parents
        self._rxd_sec_lookup = _SectionLookup()
        if sec in self._rxd_sec_lookup:
            self._rxd_sec_lookup[sec].append(self)
        else:
            self._rxd_sec_lookup[sec] = [self]

    @property
    def _sec(self):
        sl = list(self._secref)
        return None if not sl else sl[0]

    def __req__(self, other):
        if isinstance(other, nrn.Section):
            return self._sec == other
        return id(self) == id(other)

    def __rne__(self, other):
        # necessary for Python 2 but not for Python 3
        return not (self == other)

    def __eq__(self, other):
        if isinstance(other, nrn.Section):
            return self._sec == other
        return id(self) == id(other)

    def __ne__(self, other):
        # necessary for Python 2 but not for Python 3
        return not (self == other)

    def _init_diffusion_rates(self):
        # call only after roots are set
        node._diffs[self._offset : self._offset + self.nseg] = self._diff

    def _update_node_data(self):
        nseg_changed = False
        if self._sec is None:
            self._delete()
            return True
        if self._nseg != self._sec.nseg:
            num_roots = self._species()._num_roots
            offset = node._allocate(self._sec.nseg + num_roots + 1)
            replace(self, offset, self._sec.nseg)
            nseg_changed = True
        volumes, surface_area, diffs = node._get_data()
        geo = self._region._geometry
        volumes[self._offset : self._offset + self._nseg] = geo.volumes1d(self)
        surface_area[self._offset : self._offset + self._nseg] = geo.surface_areas1d(
            self
        )
        self._neighbor_areas = geo.neighbor_areas1d(self)
        if nseg_changed:
            volumes[(self._offset - self.species._num_roots) : self._offset] = 0
            self._init_diffusion_rates()
        return nseg_changed

    def _delete(self):
        # memory in node global variables is handled by the species
        _rxd_sec_lookup = _SectionLookup()

        # remove ref to this section -- at exit weakref.ref might be none
        if self._sec:
            if self._sec in _rxd_sec_lookup:
                sec_list = [s for s in _rxd_sec_lookup[self._sec] if s != self]
                if sec_list == []:
                    del _rxd_sec_lookup[self._sec]
                else:
                    _rxd_sec_lookup[self._sec] = sec_list
        else:
            _rxd_sec_lookup.remove(self)

        # check if memory has already been free'd
        if self._nseg < 0:
            return

        # node data is removed here in case references to sec remains
        if hasattr(self, "_num_roots"):
            self._nseg += self._num_roots
            self._offset -= self._num_roots
        node._remove(self._offset, self._offset + self._nseg + 1)

        # shift offset to account for deleted sec
        for secs in _rxd_sec_lookup.values():
            for s in secs:
                if s._offset > self._offset:
                    s._offset -= self._nseg + 1

        # mark memory as free'd
        self._nseg = -1

    def __del__(self):
        self._delete()

    @property
    def indices(self):
        return list(range(self._offset, self._offset + self.nseg))

    def _setup_currents(self, indices, scales, ptrs, cur_map):
        from . import rxd

        if (
            self.nrn_region is not None
            and self.species.name is not None
            and self.species.charge != 0
        ):
            ion_curr = "_ref_i%s" % self.species.name
            indices.extend(self.indices)
            volumes, surface_area, diffs = node._get_data()
            # TODO: this implicitly assumes that o and i border the membrane
            # different signs depending on if an outward current decreases the region's concentration or increases it
            if self.nrn_region == "i":
                sign = -1
            elif self.nrn_region == "o":
                sign = 1
            else:
                raise RxDException(
                    "bad nrn_region for setting up currents (should never get here)"
                )
            scales.append(
                sign
                * surface_area[self.indices]
                * 10000.0
                / (self.species.charge * h.FARADAY * volumes[self.indices])
            )
            for i in range(self.nseg):
                seg = self._sec((i + 0.5) / self.nseg)
                cur_map[self.species.name + self.nrn_region][seg] = len(ptrs)
                ptrs.append(getattr(seg, ion_curr))
            # ptrs.extend([self._sec((i + 0.5) / self.nseg).__getattribute__(ion_curr) for i in range(self.nseg)])

    @property
    def nodes(self):
        dx = self.L / self.nseg
        return nodelist.NodeList(
            [node.Node1D(self, i, ((i + 0.5) * dx) / self.L) for i in range(self.nseg)]
        )

    def _transfer_to_legacy(self):
        states = node._get_states()
        if self._concentration_ptrs is not None:
            for i, ptr in zip(
                list(range(self._offset, self._offset + self.nseg)),
                self._concentration_ptrs,
            ):
                ptr[0] = states[i]

    def _register_cptrs(self):
        global _all_cptrs, _all_cindices
        if self.nrn_region is not None and self.species.name is not None:
            ion = "_ref_" + self.species.name + self.nrn_region
            nseg = self.nseg
            for i in range(nseg):
                x = (i + 0.5) / nseg
                _all_cptrs.append(getattr(self._sec(x), ion))
                _all_cindices.append(self._offset + i)
            self._concentration_ptrs = _all_cptrs[-nseg:]
        else:
            self._concentration_ptrs = []

    def _setup_diffusion_matrix(self, mat):
        if not self._sec:
            if self._nseg > 0:
                self._delete()
            return
        _volumes, _surface_area, _diffs = node._get_data()
        offset = self._offset
        dx = self.L / self.nseg
        # print 'volumes:', _volumes
        # print 'areas:', self._neighbor_areas
        for i in range(self.nseg):
            io = i + offset
            if i > 0:
                il = io - 1
            else:
                parent, parenti = self._parent
                if isinstance(parent, weakref.ref):
                    parent = parent()
                il = parent._offset + parenti
            # second order accuracy needs diffusion constants halfway
            # between nodes, which we approx by averaging
            # TODO: is this the best way to handle boundary nodes?
            if i > 0:
                d_l = (_diffs[io] + _diffs[io - 1]) / 2.0
            else:
                d_l = _diffs[io]
            if i < self.nseg - 1:
                d_r = (_diffs[io] + _diffs[io + 1]) / 2.0
            else:
                d_r = _diffs[io]
            rate_l = d_l * self._neighbor_areas[i] / (_volumes[io] * dx)
            if i == 0:
                # on left edge, only half distance
                # TODO: verify this is the right thing to do
                rate_l *= 2
            rate_r = d_r * self._neighbor_areas[i + 1] / (_volumes[io] * dx)
            if i == self.nseg - 1:
                # on right edge, only half distance
                # TODO: verify this is the right thing to do
                rate_r *= 2
            # TODO: does this try/except add overhead? remove
            # try:
            #    g[io, io] += rate_r + rate_l
            #    g[io, io + 1] -= rate_r
            #    g[io, il] -= rate_l
            # except IndexError:
            #    print(('indexerror: g.shape = %r, io = %r, il = %r, len(node._states) = %r' % (g.shape, io, il, len(node._states))))
            #    raise
            add_values(mat, io, [il, io, io + 1], [-rate_l, rate_r + rate_l, -rate_r])

            # TODO: verify that these are correct
            if i == 0:
                # for 0 volume nodes, must conserve MASS not concentration
                scale = _volumes[io] / _volumes[il] if _volumes[il] else _volumes[io]
                add_values(mat, il, [il, io], [rate_l * scale, -rate_l * scale])
            if i == self.nseg - 1:
                ir = io + 1
                # for 0 volume nodes, must conserve MASS not concentration
                # I suspect ir always points to a 0 volume node, but
                # good to be safe
                scale = _volumes[io] / _volumes[ir] if _volumes[ir] else _volumes[io]
                add_values(mat, ir, [ir, io], [rate_r * scale, -rate_r * scale])

    def _import_concentration(self, init):
        """imports concentration from NEURON; else 0s it if not in NEURON"""
        states = node._get_states()
        if self.nrn_region is not None and self.species.name is not None:
            for i, ptr in zip(
                list(range(self._offset, self._offset + self.nseg)),
                self._concentration_ptrs,
            ):
                states[i] = ptr[0]

        elif init:
            # TODO: think about if this is the desired behavior
            for i in range(self.nseg):
                states[i + self._offset] = 0

    def _assign_parents(self, root_id, missing, root_children):
        # assign parents and root nodes
        parent_seg = _parent(self._sec)
        parent_sec = parent_seg.sec if parent_seg else None
        if parent_sec is None or not self._species()._has_region_section(
            self._region, parent_sec
        ):
            if parent_sec is None:
                pt = (self._region, None)
            else:
                pt = (self._region, parent_sec)
            if parent_sec is None or pt not in missing:
                missing[pt] = root_id
                root_id += 1
                root_children.append([])
            local_root = missing[pt]
            # TODO: which way do I want to do the first entry here?
            # self._parent = (self, local_root)
            self._parent = (self._species, local_root)
            root_children[local_root].append(self)
        else:
            # TODO: this is inefficient since checking the list twice for _region, parent_sec combo
            #       but doesn't matter much since only done at setup
            parent_section = self.species._region_section(self._region, parent_sec)
            self._parent = (parent_section, int(parent_sec.nseg * parent_seg.x))
        # print 'parent:', self._parent
        return root_id



import distutils.util
import os

# Hacky, but it's non-trivial to pass commandline arguments to pytest tests.
enable_gpu = bool(
    distutils.util.strtobool(os.environ.get("CORENRN_ENABLE_GPU", "false"))
)
mpi4py_option = bool(
    distutils.util.strtobool(os.environ.get("NRN_TEST_SPIKES_MPI4PY", "false"))
)
file_mode_option = bool(
    distutils.util.strtobool(os.environ.get("NRN_TEST_SPIKES_FILE_MODE", "false"))
)
nrnmpi_init_option = bool(
    distutils.util.strtobool(os.environ.get("NRN_TEST_SPIKES_NRNMPI_INIT", "false"))
)

# following at top level and early enough avoids...
# *** The MPI_Iprobe() function was called after MPI_FINALIZE was invoked.

# mpi4py needs to be imported before importing h
if mpi4py_option:
    from mpi4py import MPI
    from neuron import h, gui
# without mpi4py we need to call nrnmpi_init explicitly
elif nrnmpi_init_option:
    from neuron import h

    h.nrnmpi_init()
    # if mpi is active, don't ask for gui til it is turned off for all ranks > 0
    from neuron import gui
# otherwise serial execution
else:
    from neuron import h, gui


def test_spikes(
    use_mpi4py=mpi4py_option,
    use_nrnmpi_init=nrnmpi_init_option,
    file_mode=file_mode_option,
):
    print(
        "test_spikes(use_mpi4py={}, use_nrnmpi_init={}, file_mode={})".format(
            use_mpi4py, use_nrnmpi_init, file_mode
        )
    )
    h("""create soma""")
    h.soma.L = 5.6419
    h.soma.diam = 5.6419
    h.soma.insert("hh")
    h.soma.nseg = 3
    ic = h.IClamp(h.soma(0.25))
    ic.delay = 0.1
    ic.dur = 0.1
    ic.amp = 0.3

    ic2 = h.IClamp(h.soma(0.75))
    ic2.delay = 5.5
    ic2.dur = 1
    ic2.amp = 0.3

    h.tstop = 10
    h.cvode.use_fast_imem(1)
    h.cvode.cache_efficient(1)

    pc = h.ParallelContext()

    pc.set_gid2node(pc.id() + 1, pc.id())
    myobj = h.NetCon(h.soma(0.5)._ref_v, None, sec=h.soma)
    pc.cell(pc.id() + 1, myobj)

    # NEURON run
    nrn_spike_t = h.Vector()
    nrn_spike_gids = h.Vector()

    # rank 0 record spikes for all gid while others
    # for specific gid. this is for better test coverage.
    pc.spike_record(-1 if pc.id() == 0 else (pc.id() + 1), nrn_spike_t, nrn_spike_gids)

    h.run()

    nrn_spike_t = nrn_spike_t.to_python()
    nrn_spike_gids = nrn_spike_gids.to_python()

    # CORENEURON run
    from neuron import coreneuron

    coreneuron.enable = True
    coreneuron.gpu = enable_gpu
    coreneuron.file_mode = file_mode
    coreneuron.verbose = 0
    corenrn_all_spike_t = h.Vector()
    corenrn_all_spike_gids = h.Vector()

    pc.spike_record(-1, corenrn_all_spike_t, corenrn_all_spike_gids)

    pc.set_maxstep(10)

    def run(mode):
        h.stdinit()
        if mode == 0:
            pc.psolve(h.tstop)
        elif mode == 1:
            while h.t < h.tstop:
                pc.psolve(h.t + 1.0)
        else:
            while h.t < h.tstop:
                h.continuerun(h.t + 0.5)
                pc.psolve(h.t + 0.5)

        corenrn_all_spike_t_py = corenrn_all_spike_t.to_python()
        corenrn_all_spike_gids_py = corenrn_all_spike_gids.to_python()

        # check spikes match
        assert len(nrn_spike_t)  # check we've actually got spikes
        assert len(nrn_spike_t) == len(nrn_spike_gids)  # matching no. of gids
        if nrn_spike_t != corenrn_all_spike_t_py:
            print(mode)
            print(nrn_spike_t)
            print(nrn_spike_gids)
            print(corenrn_all_spike_t_py)
            print(corenrn_all_spike_gids_py)
            print(
                [
                    corenrn_all_spike_t[i] - nrn_spike_t[i]
                    for i in range(len(nrn_spike_t))
                ]
            )
        assert nrn_spike_t == corenrn_all_spike_t_py
        assert nrn_spike_gids == corenrn_all_spike_gids_py

    if file_mode is False:
        for mode in [0, 1, 2]:
            run(mode)
    else:
        run(0)

    return h


if __name__ == "__main__":
    h = test_spikes()
    if mpi4py_option or nrnmpi_init_option:
        pc = h.ParallelContext()
        pc.barrier()
    h.quit()



import distutils.util
import os

from neuron import h, gui

pc = h.ParallelContext()


def model():
    pc.gid_clear()
    for s in h.allsec():
        h.delete_section(sec=s)
    s = h.Section()
    s.L = 10
    s.diam = 10
    s.insert("hh")
    ic = h.IClamp(s(0.5))
    ic.delay = 0.1
    ic.dur = 0.1
    ic.amp = 0.5 * 0
    syn = h.ExpSyn(s(0.5))
    nc = h.NetCon(None, syn)
    nc.weight[0] = 0.001
    return {"s": s, "ic": ic, "syn": syn, "nc": nc}


def test_psolve():
    # sequence of psolve with only beginning initialization
    m = model()
    vvec = h.Vector()
    h.tstop = 5
    vvec.record(m["s"](0.5)._ref_v, sec=m["s"])

    def run(tstop):
        pc.set_maxstep(10)
        h.finitialize(-65)
        m["nc"].event(3.5)
        m["nc"].event(2.6)
        h.continuerun(1)  # Classic NEURON so psolve starts at t>0
        while h.t < tstop:
            pc.psolve(h.t + 1)

    run(h.tstop)
    vvec_std = vvec.c()  # standard result

    from neuron import coreneuron

    coreneuron.enable = True
    coreneuron.verbose = 0
    coreneuron.gpu = bool(
        distutils.util.strtobool(os.environ.get("CORENRN_ENABLE_GPU", "false"))
    )
    h.CVode().cache_efficient(True)
    run(h.tstop)
    if vvec_std.eq(vvec) == 0:
        for i, x in enumerate(vvec_std):
            print("%.3f %g %g %g" % (i * h.dt, x, vvec[i], x - vvec[i]))
    assert vvec_std.eq(vvec)
    assert vvec_std.size() == vvec.size()
    coreneuron.enable = False


def test_NetStim_noise():
    # Can use noiseFromRandom
    use_noiseFromRandom123 = False

    cells = {gid: (h.NetStim(), h.Random()) for gid in range(pc.id(), 5, pc.nhost())}
    for gid, cell in cells.items():
        pc.set_gid2node(gid, pc.id())
        pc.cell(gid, h.NetCon(cell[0], None))
        cell[1].Random123(gid, 2, 3)
        cell[1].negexp(1)
        if use_noiseFromRandom123:
            cell[0].noiseFromRandom123(gid, 2, 3)
        else:
            cell[0].noiseFromRandom(cell[1])
        cell[0].interval = gid + 1
        cell[0].number = 100
        cell[0].start = 0
        cell[0].noise = 1

    spiketime = h.Vector()
    spikegid = h.Vector()
    pc.spike_record(-1, spiketime, spikegid)

    pc.set_maxstep(10)
    tstop = 100
    for cell in cells.values():
        cell[1].seq(0)  # only _ran_compat==2 initializes the streams
    h.finitialize()
    pc.psolve(tstop)
    spiketime_std = spiketime.c()
    spikegid_std = spikegid.c()
    print("spiketime_std.size %d" % spiketime_std.size())

    spiketime.resize(0)
    spikegid.resize(0)
    from neuron import coreneuron

    h.CVode().cache_efficient(True)
    coreneuron.verbose = 0
    coreneuron.enable = True
    for cell in cells.values():
        cell[1].seq(0)
    h.finitialize()
    while h.t < tstop:
        h.continuerun(h.t + 5)
        pc.psolve(h.t + 5)

    print("spiketime.size %d" % spiketime.size())
    assert spiketime.eq(spiketime_std)
    assert spikegid.eq(spikegid_std)

    pc.gid_clear()


if __name__ == "__main__":
    test_psolve()
    test_NetStim_noise()
    for i in range(0):
        test_NetStim_noise()  # for checking memory leak
    h.quit()



import uuid
import threading
import ctypes
import json
import weakref
import time
import neuron
from neuron import h

_update_thread = None
_gui_widgets = []
_shape_plot_list = []
_active_container = None
_structure_change_count = neuron.nrn_dll_sym("structure_change_cnt", ctypes.c_int)
_diam_change_count = neuron.nrn_dll_sym("diam_change_cnt", ctypes.c_int)
_last_diam_change_count = _diam_change_count.value
_last_structure_change_count = _structure_change_count.value
update_interval = 0.1  # seconds
_has_setup = False


def _ensure_update_thread():
    global _update_thread
    if _update_thread is None:
        _update_thread = threading.Thread(target=_do_updates)
        _update_thread.daemon = True
        _update_thread.start()


def _do_updates():
    global _last_diam_change_count, _last_structure_change_count
    while True:
        old_diam_changed = h.diam_changed
        h.define_shape()
        if (
            old_diam_changed
            or h.diam_changed
            or _diam_change_count.value != _last_diam_change_count
            or _structure_change_count.value != _last_structure_change_count
        ):
            h.diam_changed = 0
            _last_diam_change_count = _diam_change_count.value
            _last_structure_change_count = _structure_change_count.value
            _do_reset_geometry()
        for obj in _gui_widgets:
            obj = obj()
            if obj is not None and obj._ready:
                obj._update()
        time.sleep(update_interval)


def _do_reset_geometry():
    h.define_shape()
    secs = list(h.allsec())
    geo = []
    for sec in secs:
        geo += _segment_3d_pts(sec)
    javascript_embedder(("set_neuron_section_data(%s);" % json.dumps(geo)))
    for sp in _shape_plot_list:
        sp = sp()
        if sp is not None:
            sp._reload_morphology()  # could probably get rid of this, but slower response
            sp._force_redraw = True
    del secs


def _ensure_setup():
    """deferred import (and webgl) loader"""
    global _has_setup, widgets
    global display, HTML, Javascript
    global setup_threejs, javascript_embedder
    global _segment_3d_pts
    global rangevars_present, _get_ptrs

    if not _has_setup:
        _has_setup = True
        import ipywidgets as widgets
        from IPython.display import display, HTML, Javascript
        from .setup_threejs import setup_threejs, javascript_embedder
        from .utilities import _segment_3d_pts
        from .rangevar import rangevars_present, _get_ptrs

        setup_threejs(
            after="""
<script>
var neuron_section_data = undefined;

function set_neuron_section_data(new_data) {
    neuron_section_data = new_data;
}

function ShapePlot(container_name) {
    this.diam_scale = 1;
    this.mode = 1;
    this.tc = new ThreeContainer(container_name);
    this.container = $('#' + container_name);
    this.section_data = undefined;
    this.vmin = -100;
    this.vmax = 100;
}

ShapePlot.prototype.update = function() {
    if (this.section_data !== neuron_section_data) {
        this.section_data = neuron_section_data;
        this.tc.clearLines();
        var my_mode = this.mode;
        function const_function(val) {
            return function(p) {return val};
        }
        function sp_interpolater(arc3d, ys, diam_scale) {
            var my_arc3d = [];
            var lo = arc3d[0];
            var hi = arc3d[arc3d.length - 1];
            var delta = hi - lo;
            for (var i = 0; i < arc3d.length; i++) {
                my_arc3d.push((arc3d[i] - lo) / delta);
            }
            return function(p) {
                for(var i = 1; i < my_arc3d.length; i++) {
                    var me = my_arc3d[i];
                    if (p < me) {
                        var last = my_arc3d[i - 1];
                        var x = (p - last) / (me - last);
                        return diam_scale * (x * ys[i] + (1 - x) * ys[i - 1]);
                    }
                }
                return ys[my_arc3d.length - 1] * diam_scale;
            }
        }
        var const_diam_f = const_function(4 * this.diam_scale);
        var my_width_rule;
        for(var i = 0; i < this.section_data.length; i++) {
            var my_segment = this.section_data[i];
            var xs = my_segment[0];
            var ys = my_segment[1];
            var zs = my_segment[2];
            var ds = my_segment[3];
            var arcs = my_segment[4];
            var geo = new THREE.Geometry();
            for(var j = 0 ; j < xs.length; j++) {
                geo.vertices.push(new THREE.Vector3(xs[j], ys[j], zs[j]));
            }
            if (this.mode == 0) {
                my_width_rule = sp_interpolater(arcs, ds, 4 * this.diam_scale);
            } else {
                my_width_rule = const_diam_f;
            }
            this.tc.makeLine(geo, my_width_rule);
        }
    }
}

ShapePlot.prototype.force_update = function() {
    this.section_data = undefined;
    this.update();
}


ShapePlot.prototype.set_diam_scale = function(diam) {
    this.diam_scale = diam;
    this.force_update();
}

ShapePlot.prototype.update_colors = function(data) {
    var vmin = this.vmin;
    var vmax = this.vmax;
    var vdelta = vmax - vmin;
    for (var i = 0; i < this.section_data.length; i++) {
        var v = data[i];
        var r, g, b;
        if (v == null) {
            r = 0;
            g = 0;
            b = 0;
        } else {
            v = (data[i] - vmin) / vdelta;
            if (v < 0) {v = 0;}
            if (v > 1) {v = 1;}
            r = v;
            b = 1 - v;
            g = 0;
        }
        var cv = this.tc.lines[i].material.uniforms.color.value;
        cv.r = r;
        cv.g = g;
        cv.b = b;
    }
}
</script>
"""
        )


class GUIWidget:
    def __init__(self, parent=None):
        _ensure_setup()
        self._ready = False
        self._index = None
        if parent is None:
            parent = _active_container
        if parent is None:
            # parent = widgets.VBox([])
            # display(parent)
            self._index = 0
        self._parent = parent
        _ensure_update_thread()
        _gui_widgets.append(weakref.ref(self))


class PlotShape(GUIWidget):
    def __init__(self):
        from .config import options

        if options["backend"] != "jupyter":
            raise Exception(
                'gui2.PlotShape currently only works with "jupyter" backend'
            )
        GUIWidget.__init__(self)
        self._uuid = str(uuid.uuid4()).replace("-", "")
        self._graph_div = widgets.Output()
        with self._graph_div:
            display(
                HTML(
                    '<div id="%s" style="width:400px; height:400px; border: 1px solid black"></div>'
                    % self._uuid
                )
            )

        self._rangevars_control = widgets.Select(options=["No coloration"], index=0)
        parent = widgets.HBox([self._graph_div, self._rangevars_control])
        display(parent)

        self._rangevars = []
        self._val = []
        self._ptrs = []
        self._old_selection = None

        display(
            HTML(
                '<script>sp%s = new ShapePlot("%s");</script>'
                % (self._uuid, self._uuid)
            )
        )
        _shape_plot_list.append(weakref.ref(self))
        self._min = -100
        self._max = 100
        self._force_redraw = False
        self._update_colors = 5
        self._ready = True

    def camera_position(self, x, y, z):
        """set the camera position"""
        javascript_embedder(
            "sp%s.tc.camera.position.set(%g, %g, %g)" % (self._uuid, x, y, z)
        )

    def _reload_morphology(self):
        javascript_embedder(("sp%s.update()" % self._uuid))

    def scale(self, low, high):
        """Sets blue (low) and red (high) values for the color scale."""
        javascript_embedder(
            "sp%s.vmin = %g; sp%s.vmax = %g;" % (self._uuid, low, self._uuid, high)
        )
        javascript_embedder(
            "sp%s.update_colors(%s)" % (self._uuid, json.dumps(self._val))
        )

    def _do_rangevar_update(self):
        rangevars = rangevars_present(list(h.allsec()))
        if len(rangevars) != len(self._rangevars) or any(
            r1 != r2 for r1, r2 in zip(rangevars, self._rangevars)
        ):
            self._rangevars = rangevars
            self._rangevars_control.options = [r["name"] for r in rangevars]
            for i, rv in enumerate(rangevars):
                if rv["name"] == self._old_selection:
                    self._rangevars_control.index = i
                    break
            self._force_redraw = True

    def _update(self):
        if not self._ready:
            return
        self._do_rangevar_update()
        if (
            self._rangevars_control.options[self._rangevars_control.index]
            != self._old_selection
            or self._force_redraw
        ):
            self._old_selection = self._rangevars_control.options[
                self._rangevars_control.index
            ]
            self._ptrs = _get_ptrs(self._rangevars[self._rangevars_control.index])
        data = [ptr[0] if ptr is not None else None for ptr in self._ptrs]
        if (
            self._val is None
            or len(self._val) != len(data)
            or any(myval != oldval for myval, oldval in zip(data, self._val))
        ):
            if self._val is not None or any(data) is not None:
                self._update_colors += 1
            self._val = data
        self._old_selection = self._rangevars_control.options[
            self._rangevars_control.index
        ]
        if self._force_redraw:
            self._reload_morphology()
        if self._update_colors or self._force_redraw:
            javascript_embedder(
                ("sp%s.update_colors(%s)" % (self._uuid, json.dumps(self._val)))
            )
            self._update_colors = min(20, max(0, self._update_colors - 1))
        self._force_redraw = False

    def variable(self, what):
        """Sets the range variable (v, ca[cyt], hh.m, etc...) to be used for the shape plot."""
        self._do_rangevar_update()
        for i, rv in enumerate(self._rangevars):
            if rv["name"] == what:
                self._rangevars_control.index = i
                self._force_redraw = True
                self._update_colors += 5
                break
        else:
            raise Exception("No such variable " + str(what))

    def show(self, m):
        """Sets the view mode m.

        If m == 0, then displays diameter.
        If m == 1, then displays centroid.
        """
        if m not in (0, 1):
            raise Exception("view mode must be 0 or 1")
        javascript_embedder(
            ("sp%s.mode=%g; sp%s.update()" % (self._uuid, m, self._uuid))
        )
        h.diam_changed = 1

    def set_constant_diameter(self, diam):
        """sets the diameter used for mode == 1 (i.e. constant diameter/centroid view)"""
        javascript_embedder(("sp%s.set_default_diameter(%g)" % (self._uuid, diam)))



from neuron import h, crxd as rxd, gui
import numpy
import sys
import time
import itertools

npar = len(sys.argv)
# if(npar<2):
#    print "usage: python wave1d.py <nseg> <nsubseg>"
#    sys.exit(0)

# rxd.options.nsubseg =int(sys.argv[2])
rxd.options.subseg_interpolation = 0
rxd.options.subseg_averaging = 0


sec = h.Section()
sec.L = 100
sec.diam = 1
sec.nseg = 25
# h.pt3dadd(0, 0, 0, 1)
# h.pt3dadd(100, 0, 0, 1)
sec.Ra = 150
Rm = 25370
dend = sec
for myseg in dend:
    myseg.v = -64
for myseg in dend:
    myseg.cm = 1.41
dend.insert("pas")
for myseg in dend:
    myseg.pas.g = 1.0 / Rm
for myseg in dend:
    myseg.pas.e = -64
dend.insert("cal")  # insert L-type Ca channel
for myseg in dend:
    myseg.cal.gcalbar = 1.0e-6

nsubseg = 5

nstim = 5
st_dur = 2
st_interv = 50
st_start = 1000
stims = []
for i in range(nstim):
    stim = h.IClamp(0.5, sec=sec)
    stim.delay = st_start + i * (st_interv)
    stim.dur = st_dur
    stim.amp = 1.0 / 8000 * 1200
    stims.append(stim)
v1 = h.Vector()
v1.record(sec(0.5)._ref_v)
ca1 = h.Vector()
ca1.record(sec(0.5)._ref_cai)
v2 = h.Vector()
v2.record(sec(0.25)._ref_v)
ca2 = h.Vector()
ca2.record(sec(0.25)._ref_cai)
times = h.Vector()
times.record(h._ref_t)
h.finitialize()
print("baba")

s = h.PlotShape()

s.exec_menu("Shape Plot")
if __name__ == "__main__":

    def set_plotshape_colormap(plotshape, cmap="jet"):
        import matplotlib.cm

        s = matplotlib.cm.ScalarMappable(cmap=cmap)
        cmap = s.get_cmap()
        s.set_clim(0, cmap.N)
        rs, gs, bs = itertools.islice(zip(*s.to_rgba(list(range(cmap.N)))), 0, 3)
        plotshape.colormap(cmap.N)
        for i, r, g, b in zip(range(cmap.N), rs, gs, bs):
            plotshape.colormap(i, r * 255, g * 255, b * 255)
        # call s.scale(lo, hi) to replot the legend

    set_plotshape_colormap(s)

# show the diameters
s.show(0)

# cytoplasmic, er volume fractions
fc, fe = 0.83, 0.17

# parameters
caDiff = 0.016
# caDiff =0
ip3Diff = 0.283
# ip3Diff = 0
cac_init = 1.0e-4
ip3_init = 0.1
gip3r = 12040
gserca = 0.3913
gleak = 6.020
kserca = 0.1
kip3 = 0.15
kact = 0.4
ip3rtau = 2000
h.CVode().active(1)
h.cvode.atol(1e-10)

# define the regions for the rxd
cyt = rxd.Region(
    h.allsec(), nrn_region="i", geometry=rxd.FractionalVolume(fc, surface_fraction=1)
)
er = rxd.Region(h.allsec(), geometry=rxd.FractionalVolume(fe))
cyt_er_membrane = rxd.Region(h.allsec(), geometry=rxd.FixedPerimeter(1))

# the species and other states
ca = rxd.Species([cyt, er], d=caDiff, name="ca", charge=2, initial=cac_init)
ip3 = rxd.Species(cyt, d=ip3Diff, initial=ip3_init)
ip3r_gate_state = rxd.State(cyt_er_membrane, initial=0.8)
h_gate = ip3r_gate_state[cyt_er_membrane]


# pumps and channels between ER and Cytosol

serca = rxd.MultiCompartmentReaction(
    ca[cyt] > ca[er],
    gserca / ((kserca / (1000.0 * ca[cyt])) ** 2 + 1),
    membrane=cyt_er_membrane,
    custom_dynamics=True,
)
leak = rxd.MultiCompartmentReaction(
    ca[er] != ca[cyt], gleak, gleak, membrane=cyt_er_membrane
)

minf = ip3[cyt] * 1000.0 * ca[cyt] / (ip3[cyt] + kip3) / (1000.0 * ca[cyt] + kact)
k = gip3r * (minf * h_gate) ** 3
ip3r = rxd.MultiCompartmentReaction(ca[er] != ca[cyt], k, k, membrane=cyt_er_membrane)
ip3rg = rxd.Rate(h_gate, (1.0 / (1 + 1000.0 * ca[cyt] / (0.3)) - h_gate) / ip3rtau)


cae_init = (0.0017 - cac_init * fc) / fe
ca[er].concentration = cae_init

# ip3.nodes.concentration = 2
for node in ip3.nodes:
    if node.x < 0.2:
        node.concentration = 2


h.CVode().re_init()

s.variable("cai")
# s.scale(-70, -50)
s.scale(0, 2e-3)


tstop = 3000
recdt = 100
datacol = 0
tstop = 1500
del s
h.continuerun(tstop)



from neuron import h, rxd
from neuron.units import mV, ms

h.load_file("stdrun.hoc")

soma = h.Section(name="soma")
soma.diam = soma.L = 10

# the geometry of the ecs doesn't matter in the constant case
ecs = rxd.Region([soma], name="ecs", nrn_region="o", geometry=rxd.Shell(1, 2))
cyt = rxd.Region([soma], name="cyt", nrn_region="i")
mem = rxd.Region([soma], name="mem", geometry=rxd.membrane())

cao = rxd.Parameter(ecs, charge=2)

ca = rxd.Species(
    [cyt],
    d=1,  # with single section and nseg=1 only affects extracellular
    name="ca",
    charge=2,
    initial=lambda node: 1e-3 if node.region == cyt else 0,
)

e = 1.60217662e-19
scale = 1e-14 / e

# rate constant is in terms of molecules/um2 ms
ca_pump = rxd.MultiCompartmentReaction(
    ca[cyt],
    cao[ecs],
    ca[cyt] * scale,
    custom_dynamics=True,
    membrane_flux=True,
    membrane=mem,
)

t = h.Vector().record(h._ref_t)
ca_vec = h.Vector().record(soma(0.5)._ref_cai)
ca_vec2 = h.Vector().record(cao.nodes(soma(0.5))._ref_value)
v = h.Vector().record(soma(0.5)._ref_v)

h.finitialize(-65 * mV)
h.continuerun(100 * ms)

if __name__ == "__main__":
    import matplotlib.pyplot as plt

    plt.subplot(2, 1, 1)
    plt.plot(t, ca_vec * 1000, label="inside")
    plt.plot(t, ca_vec2 * 1000, label="outside")
    plt.ylabel("[Ca] (uM)")
    plt.legend()
    plt.subplot(2, 1, 2)
    plt.plot(t, v)
    plt.ylabel("v (mV)")
    plt.xlabel("t (ms)")
    plt.show()



import os
import pkgutil
import re
import sphinx

"""
This script generates a HOC domain from the one available in the sphinx package and writes it to:
    
    docs/domains/hocdomain.py

Ideally we'd have a proper HOC domain, but that is an extra workload and we lack the required knowledge to build one.
See https://github.com/neuronsimulator/nrn/issues/1540 


"""


def generate_hoc_domain_from_py():
    # Include Sphinx version for robustness
    gen_from = "# generated from 'sphinx/domains/python.py' @ Sphinx {}\n".format(
        sphinx.__version__
    )
    print("Hacking docs/domains/hocdomain.py " + gen_from)

    # Retrieve Python domain implementation
    py_domain = gen_from + pkgutil.get_data("sphinx.domains", "python.py").decode(
        "utf-8"
    )

    # regex replacements from Python to HOC domain (working Sphinx version 4.4.0)
    hoc_dict = {
        # Strings used for indexing.
        # For future updates, tthe generated index must be visually checked and x-reffed against previous index
        " (%s ": " (HOC %s ",
        "'%s (": "'%s (HOC ",
        "'built-in function'": "'HOC built-in function'",
        "(built-in variable)": "(HOC built-in variable)",
        # for Sphinx internals, basically renaming Python names of objects, types, clasess ..
        "'python": "'hoc",
        "python_": "hoc_",
        "Python": "HOC",
        "Py": "HOC",
        "'py": "'hoc",
        "py_": "hoc_",
    }

    # Create a regular expression  from the dictionary keys
    regex = re.compile("(%s)" % "|".join(map(re.escape, hoc_dict.keys())))

    # For each match, look-up corresponding value in dictionary
    hoc_domain = regex.sub(
        lambda mo: hoc_dict[mo.string[mo.start() : mo.end()]], py_domain
    )

    # Write out HOC domain
    with open(
        os.path.join(os.path.abspath("."), "domains/hocdomain.py"), "w"
    ) as hoc_py:
        hoc_py.writelines(hoc_domain)

    print("Done.")


if __name__ == "__main__":
    generate_hoc_domain_from_py()



from . import graphicsPrimitives as graphics
from .. import options
import math
import random


def get_verts(voxel, g):
    """return list (len=8) of point coordinates (x,y,z) that are vertices of the voxel (i,j,k)"""
    (i, j, k) = voxel
    dx, dy, dz = g["dx"], g["dy"], g["dz"]
    v1_0, v1_1, v1_2 = g["xlo"] + i * dx, g["ylo"] + j * dy, g["zlo"] + k * dz
    vertices = [
        (v1_0, v1_1, v1_2),
        (v1_0 + dx, v1_1, v1_2),
        (v1_0 + dx, v1_1 + dy, v1_2),
        (v1_0, v1_1 + dy, v1_2),
        (v1_0, v1_1, v1_2 + dz),
        (v1_0 + dx, v1_1, v1_2 + dz),
        (v1_0 + dx, v1_1 + dy, v1_2 + dz),
        (v1_0, v1_1 + dy, v1_2 + dz),
    ]
    return vertices


def get_subverts(voxel, g, step):
    """return list (len=8) of point coordinates (x,y,z) that are vertices of the voxel (i,j,k)"""
    (i, j, k) = voxel
    dx, dy, dz = g["dx"], g["dy"], g["dz"]
    DX, DY, DZ = step
    v1_0, v1_1, v1_2 = g["xlo"] + i * dx, g["ylo"] + j * dy, g["zlo"] + k * dz
    vertices = [
        (v1_0, v1_1, v1_2),
        (v1_0 + DX, v1_1, v1_2),
        (v1_0 + DX, v1_1 + DY, v1_2),
        (v1_0, v1_1 + DY, v1_2),
        (v1_0, v1_1, v1_2 + DZ),
        (v1_0 + DX, v1_1, v1_2 + DZ),
        (v1_0 + DX, v1_1 + DY, v1_2 + DZ),
        (v1_0, v1_1 + DY, v1_2 + DZ),
    ]
    return vertices


def add_res(flist, voxel, verts_in, res, g):
    dx, dy, dz = g["dx"], g["dy"], g["dz"]
    Sx, Sy, Sz = dx / res, dy / res, dz / res
    bit = Sx * Sy * Sz

    step = [dx / 2, dy / 2, dz / 2]
    subverts = get_subverts(
        voxel, g, step
    )  # the 'voxel positions' for the new subvoxels

    count = 0
    # select only the subvoxels of the vertices that are in
    for i in verts_in:
        v0 = subverts[i]
        startpt = (
            v0[0] + dx / (2 * res),
            v0[1] + dy / (2 * res),
            v0[2] + dx / (2 * res),
        )
        for x in range(res // 2):
            for y in range(res // 2):
                for z in range(res // 2):
                    v = (
                        startpt[0] + x * (dx / res),
                        startpt[1] + y * (dy / res),
                        startpt[2] + z * (dz / res),
                    )
                    if any(
                        f.distance(v[0], v[1], v[2]) <= options.ics_distance_threshold
                        for f in flist
                    ):
                        count += 1

    if count > 0:
        return count * bit
    return bit / 8


def Put(flist, voxel, v0, verts_in, res, g):
    """add voxel key with partial volume value to dict of surface voxels"""
    # v0 is the start coordinates of voxel (verts[0] feed in)
    # res is resolution of sampling points (only works for even values of res!)
    dx, dy, dz = g["dx"], g["dy"], g["dz"]
    Sx, Sy, Sz = dx / res, dy / res, dz / res

    count = 0
    startpt = v0[0] + Sx / 2.0, v0[1] + Sy / 2.0, v0[2] + Sz / 2.0
    for i in range(res):
        for j in range(res):
            for k in range(res):
                v = startpt[0] + i * Sx, startpt[1] + j * Sy, startpt[2] + k * Sz
                if any(
                    f.distance(v[0], v[1], v[2]) <= options.ics_distance_threshold
                    for f in flist
                ):
                    count += 1
    if count > 0:
        return Sx * Sy * Sz * count
    return add_res(flist, voxel, verts_in, 2 * res, g)


def simplevolume(flist, distances, voxel, g):
    """estimate the volume by the fraction of interior points"""
    res = options.ics_partial_volume_resolution
    distance_threshold = options.ics_distance_threshold

    sx, sy, sz = g["dx"] / res, g["dy"] / res, g["dz"] / res

    res1 = res + 1

    # count up the interior points on the boundary
    if res == 1:
        count = sum(1 if d <= distance_threshold else 0 for d in distances)

    else:
        vi, vj, vk = voxel
        startx, starty, startz = (
            g["xlo"] + vi * g["dx"],
            g["ylo"] + vj * g["dy"],
            g["zlo"] + vk * g["dz"],
        )

        count = 0
        for i in range(res1):
            vx = startx + i * sx
            for j in range(res1):
                vy = starty + j * sy
                for k in range(res1):
                    vz = startz + k * sz
                    if any(f.distance(vx, vy, vz) <= distance_threshold for f in flist):
                        count += 1
    volume = count * g["dx"] * g["dy"] * g["dz"] / (res1**3)
    return volume



import sys
from io import StringIO
import inspect
from neuron import h

pc = h.ParallelContext()
nhost = pc.nhost()

quiet = True


def set_quiet(b):
    global quiet
    old = quiet
    quiet = b
    return old


def printerr(e):
    if not quiet:
        print(e)


def checking(s):
    if not quiet:
        print("CHECKING: " + s)


def expect_hocerr(callable, args, sec=None):
    """
    Execute callable(args) and assert that it generated an error.

    If sec is not None, executes callable(args, sec=sec)
    Skips if nhost > 1 as all hoc_execerror end in MPI_ABORT
    Does not work well with nrniv launch since hoc_execerror messages do not
    pass through sys.stderr.
    """

    if nhost > 1:
        return

    original_stderr = sys.stderr
    sys.stderr = my_stderr = StringIO()
    err = 0
    pyerrmes = False
    try:
        if sec:
            callable(*args, sec=sec)
        else:
            callable(*args)
            printerr("expect_hocerr: no err for %s%s" % (str(callable), str(args)))
    except Exception as e:
        err = 1
        errmes = my_stderr.getvalue()
        if errmes:
            errmes = errmes.splitlines()[0]
            errmes = errmes[(errmes.find(":") + 2) :]
            printerr("expect_hocerr: %s" % errmes)
        elif e:
            printerr(e)
    finally:
        sys.stderr = original_stderr
    assert err


def expect_err(stmt):
    """
    expect_err('stmt')
    stmt is expected to raise an error
    """
    here = inspect.currentframe()
    caller = here.f_back
    err = 0
    checking(stmt)
    try:
        exec(stmt, caller.f_globals, caller.f_locals)
        printerr("expect_err: no err for-- " + stmt)
    except Exception as e:
        err = 1
        printerr(e)
    assert err



from neuron import h

h("create a, b")
for sec in h.allsec():
    sec.nseg = 5
    sec.insert("t2")

for seg in h.a:
    seg.t2.r = 10 + seg.x * 9
    h.b(seg.x).t2.r = 40 + seg.x * 9
    # h.setpointer(h.b(seg.x)._ref_r_t2, 'p', seg.t2)
    # h.setpointer(seg._ref_r_t2, 'p', h.b(seg.x).t2)
    seg.t2._ref_p = h.b(seg.x)._ref_r_t2
    h.b(seg.x)._ref_p_t2 = seg.t2._ref_r


def pr(sec):
    sec.push()
    for seg in sec:
        print(
            "%s   x=%g   r=%g   t2.p=%g   p_t2=%g"
            % (sec.name(), seg.x, seg.t2.r, seg.t2.p, seg.p_t2)
        )
        h.setdata_t2(seg.x)
        print("   f()=%g" % h.f_t2())
    print("\n")
    h.pop_section()


pr(h.a)
pr(h.b)

z = h.a(0.5).t2
print(z)
print(z._ref_p)
z._ref_p[0] = 5
print(h.b(0.5).t2.r)
z.p = 10
print(h.b(0.5).t2.r)



from .rxdException import RxDException
import weakref
from . import species, rxdmath, rxd, initializer
import numpy
from .rangevar import RangeVar
import itertools
import warnings
from .generalizedReaction import GeneralizedReaction

# aliases to avoid repeatedly doing multiple hash-table lookups
_itertools_chain = itertools.chain


class Rate(GeneralizedReaction):
    """Declare a contribution to the rate of change of a species or other state variable.

    Example:

        constant_production = rxd.Rate(protein, k)

    If this was the only contribution to protein dynamics and there was no
    diffusion, the above would be equivalent to:

        dprotein/dt = k

    If there are multiple rxd.Rate objects (or an rxd.Reaction, etc) acting on
    the same species, then their effects are summed.
    """

    def __init__(self, species, rate, regions=None, membrane_flux=False):
        """create a rate of change for a species on a given region or set of regions

        if regions is None, then does it on all regions"""
        self._species = weakref.ref(species)
        self._original_rate = rate
        from . import region

        if not hasattr(regions, "__len__"):
            if regions is not None and not isinstance(regions, region.Extracellular):
                regions = [regions]
            else:
                if hasattr(species, "_extracellular"):
                    regions = [species._extracellular()._region]
                elif hasattr(species, "region"):
                    regions = [species._region()]
                else:
                    regions = []
                    if hasattr(species, "_extracellular_regions"):
                        regions += species._extracellular_regions
                    if hasattr(species, "_regions"):
                        regions += species._regions
                    if not regions:
                        regions = [None]
        else:
            regions = [reg for reg in regions]
        self._regions = regions
        self._membrane_flux = membrane_flux
        if membrane_flux not in (True, False):
            raise RxDException("membrane_flux must be either True or False")
        if membrane_flux and regions is None:
            # TODO: rename regions to region?
            raise RxDException(
                "if membrane_flux then must specify the (unique) membrane regions"
            )
        self._trans_membrane = False
        rxd._register_reaction(self)

        # be careful, this could keep states alive
        self._original_rate = rate

        try:
            self._voltage_dependent = rate._voltage_dependent
        except AttributeError:
            self._voltage_dependent = False

        # initialize self if the rest of rxd is already initialized
        if initializer.is_initialized():
            self._do_init()

    def _do_init(self):
        from . import region, species

        rate = self._original_rate
        if not isinstance(rate, RangeVar):
            regions = []
            if self._regions and self._regions != [None]:
                regions = self._regions
            elif hasattr(self._species(), "_regions"):
                regions = self._species()._regions
            elif hasattr(self._species(), "_region"):
                regions = [self._species()._region()]
            regions = [
                r
                for r in regions
                if isinstance(r, region.Extracellular) or (r._secs1d or r._secs3d)
            ]
            if regions:
                self._rate, self._involved_species = rxdmath._compile(rate, regions)
            else:
                self._rate, self._involved_species = {}, []
        else:
            self._involved_species = self._species
        self._update_indices()

        # Check to if it is an extracellular reaction
        # Was an ECS region was passed to to the constructor
        ecs_region = [r for r in self._regions if isinstance(r, region.Extracellular)]
        # ecs_region = ecs_region[0] if len(ecs_region) > 0 else None
        # commented above out line because we need to pass all ecs_regions in.
        # TODO review this code below...I am very tired. Also did this in reaction.py
        ecs_region = ecs_region if len(ecs_region) > 0 else None
        if ecs_region:
            self._ecs_regions = ecs_region
        # Is the species passed to the constructor extracellular
        if not ecs_region:
            if isinstance(self._species(), species.SpeciesOnExtracellular):
                ecs_region = self._species()._extracellular()
            elif isinstance(self._species(), species._ExtracellularSpecies):
                ecs_region = self._species()._region

        # Is the species passed to the constructor defined on the ECS
        if not ecs_region:
            if not isinstance(self._species(), species.SpeciesOnRegion):
                sp = self._species()
                if sp and sp._extracellular_instances:
                    self._ecs_regions = [
                        key for key in sp._extracellular_instances.keys()
                    ]

        if hasattr(self, "_ecs_regions"):
            self._rate_ecs, self._involved_species_ecs = rxdmath._compile(
                rate, self._ecs_regions
            )
            for sp in self._involved_species_ecs:
                if ecs_region and isinstance(sp(), species.SpeciesOnRegion):
                    raise RxDException(
                        "Error Rate %r: an extracellular rate can not depend on a SpeciesOnRegion"
                        % self._original_rate
                    )

    def __repr__(self):
        short_rate = (
            self._original_rate._short_repr()
            if hasattr(self._original_rate, "_short_repr")
            else self._original_rate
        )
        if len(self._regions) != 1 or self._regions[0] is not None:
            regions_short = (
                "[" + ", ".join(r._short_repr() for r in self._regions) + "]"
            )
            return "Rate(%s, %s, regions=%s, membrane_flux=%r)" % (
                self._species()._short_repr(),
                short_rate,
                regions_short,
                self._membrane_flux,
            )
        else:
            return "Rate(%s, %s, membrane_flux=%r)" % (
                self._species()._short_repr(),
                short_rate,
                self._membrane_flux,
            )

    def _rate_from_rangevar(self, *args):
        return self._original_rate._rangevar_vec()

    def _update_indices(self):
        # this is called anytime the geometry changes as well as at init
        # TODO: is the above statement true?

        # Default values
        self._indices_dict = {}
        self._indices = []
        self._jac_rows = []
        self._jac_cols = []
        self._mult = [1]
        self._mult_extended = self._mult
        from . import region

        if (
            not self._species()
            or isinstance(self._species(), species.SpeciesOnExtracellular)
            or all([isinstance(r, region.Extracellular) for r in self._regions])
        ):
            if self._regions != [None]:
                self._active_regions = self._regions
            elif self._species():
                if isinstance(self._species(), species.SpeciesOnExtracellular):
                    self._active_regions = [self._species()._extracellular()._region]
            if hasattr(self, "_involved_species"):
                for sptr in self._involved_species:
                    self._indices_dict[sptr()] = []
            return

        active_secs = None

        # locate the regions containing all species (including the one that changes)
        active_regions = list(
            set.intersection(
                *[
                    set(
                        sptr()._regions + sptr()._extracellular_regions
                        if isinstance(sptr(), species.Species)
                        else [
                            sptr()._region()
                            if isinstance(sptr(), species.SpeciesOnRegion)
                            else sptr()._extracellular()
                        ]
                    )
                    for sptr in list(self._involved_species) + [self._species]
                ]
            )
        )
        sp_regions = (
            self._species()._regions + self._species()._extracellular_regions
            if isinstance(self._species(), species.Species)
            else [
                self._species()._region()
                if isinstance(self._species(), species.SpeciesOnRegion)
                else self._species()._extracellular()
            ]
        )
        actr = sp_regions
        if self._regions != [None]:
            specified_regions = list(
                set.intersection(set(self._regions), set(sp_regions))
            )
            if specified_regions:
                actr = specified_regions
                active_regions = list(set.intersection(set(active_regions), set(actr)))
            else:
                warnings.warn(
                    "Error in rate %r\nThe regions specified %s are not appropriate regions %s will be used instead."
                    % (
                        self,
                        [r._name for r in self._regions],
                        [r._name for r in self._species()._regions],
                    )
                )

        # Note: This finds the sections where the all involved species exists
        # e.g. for Rate(A, B+C) if A sections [1,2,3] and B is on sections [1,2] and C is on sections [2,3]
        # The Rate will only effect A on section 2 (rather than have 3 different rates)
        if not active_regions:  # They do not share a common region
            if any(
                isinstance(sptr(), species.Species)
                for sptr in list(self._involved_species) + [self._species]
            ):
                for sptr in self._involved_species:
                    self._indices_dict[sptr()] = []
                return
            active_1d_secs = list(
                set.intersection(
                    *[set(reg._secs1d) for reg in actr if hasattr(reg, "_secs1d")]
                )
            )
            active_3d_secs = list(
                set.intersection(
                    *[set(reg._secs3d) for reg in actr if hasattr(reg, "_secs3d")]
                )
            )
            active_regions = actr
            # if there are multiple regions on a segment for an involved species the rate is ambiguous
            for sptr in self._involved_species:
                s = sptr()
                indices = [list(s.indices(secs={sec})) for sec in active_1d_secs]
                if not all(
                    rcount == sec.nseg or rcount == 0
                    for rcount, sec in zip(
                        [len(ind) for ind in indices], active_1d_secs
                    )
                ):
                    raise RxDException(
                        "Error in rate %r, the species do not share a common region"
                        % self
                    )
                # remove sections where species is absent
                active_1d_secs = {
                    sec
                    for sec, ind in zip(active_1d_secs, indices)
                    if len(ind) == sec.nseg
                }
            # Repeated with the trimmed active_1d_secs and store the indices
            if active_1d_secs:
                for sptr in self._involved_species:
                    s = sptr()
                    indices = [list(s.indices(secs={sec})) for sec in active_1d_secs]
                    self._indices_dict[s] = list(
                        _itertools_chain.from_iterable(indices)
                    )
                    self._indices = [self._species().indices(actr, active_1d_secs)]
            elif active_3d_secs:
                pass
                # Account for hybrid stuff as well
                # print("we have an active 3d sec")
            else:
                raise RxDException(
                    "Error in rate %r, the species do not share a common section" % self
                )
        else:
            active_secs = [
                set(reg.secs) for reg in active_regions if hasattr(reg, "secs")
            ]
            if active_secs:
                active_secs = set.union(*active_secs)
            else:
                active_secs = set()
            # store the indices
            for sptr in self._involved_species:
                s = sptr()
                self._indices_dict[s] = s.indices(active_regions, active_secs)

            self._indices = [self._species().indices(active_regions, active_secs)]
        self._active_regions = active_regions

        if isinstance(self._original_rate, RangeVar):
            nodes = []
            for sptr in self._involved_species:
                s = sptr()
                for r in active_regions:
                    if r is None:
                        nodes += s.nodes
                    else:
                        nodes += s[r].nodes
            self._original_rate._init_ptr_vectors(nodes)
            self._indices = [self._original_rate._locs]
            self._rate = self._rate_from_rangevar
            self._get_args = lambda ignore: []
        else:
            self._update_jac_cache()

    def _do_memb_scales(self):
        # TODO: does anyone still call this?
        # TODO: update self._memb_scales (this is just a dummy value to make things run)
        self._memb_scales = 1

    def _get_memb_flux(self, states):
        if self._membrane_flux:
            # raise RxDException('membrane flux due to rxd.Rate objects not yet supported')
            # TODO: refactor the inside of _evaluate so can construct args in a separate function and just get self._rate() result
            rates = self._evaluate(states)[2]
            return self._memb_scales * rates
        else:
            return []



def test_volumes_areas(neuron_instance):
    """Mode to test the 3D vocalization areas and volumes"""
    h, rxd, data, save_path = neuron_instance
    dend = h.Section(name="dend")
    dend.L = 2
    dend.diam = 2
    dend.nseg = 9
    dx = 0.25

    rxd.set_solve_type(dimension=3)
    cyt = rxd.Region([dend], name="cyt", dx=dx)
    ca = rxd.Species(cyt, name="ca", charge=2)

    nodes_on_surface = [node for node in ca.nodes if node.surface_area]
    nodes_internal = [node for node in ca.nodes if not node.surface_area]

    correct_surface_area = 15.713551302745485
    correct_surface_vol = 2.398148148148151
    correct_internal_vol = 3.0
    correct_vol = 5.398148148148163

    surface_area = sum([nd.surface_area for nd in nodes_on_surface])
    surface_vol = sum([nd.volume for nd in nodes_on_surface])
    internal_vol = sum([nd.volume for nd in nodes_internal])
    vol = sum([nd.volume for nd in ca.nodes])

    assert abs(correct_surface_area - surface_area) / correct_surface_area < 0.01
    assert abs(correct_surface_vol - surface_vol) / correct_surface_vol < 0.01
    assert abs(correct_internal_vol - internal_vol) / correct_internal_vol < 0.01
    assert abs(correct_vol - vol) / correct_vol < 0.01



import re
import sys
from neuron.expect_hocerr import expect_hocerr, expect_err, set_quiet

import numpy as np

from neuron import h, hoc


def test_soma():
    h("""create soma""")

    assert h.soma is not None

    h.soma.L = 5.6419
    h.soma.diam = 5.6419

    assert h.soma.L == 5.6419
    assert h.soma.diam == 5.6419


def test_simple_sim():
    h("""create soma""")
    h.load_file("stdrun.hoc")
    h.soma.L = 5.6419
    h.soma.diam = 5.6419
    expect_hocerr(h.ion_register, ("na", 2))
    assert h.ion_charge("na_ion") == 1.0
    expect_hocerr(h.ion_register, ("ca", 3))
    assert h.ion_charge("ca_ion") == 2.0
    h.soma.insert("hh")
    ic = h.IClamp(h.soma(0.5))
    ic.delay = 0.5
    ic.dur = 0.1
    ic.amp = 0.3

    v = h.Vector()
    v.record(h.soma(0.5)._ref_v, sec=h.soma)
    tv = h.Vector()
    tv.record(h._ref_t, sec=h.soma)
    h.run()
    assert v[0] == -65.0


def test_spikes():
    ref_spikes = np.array([1.05])

    h("""create soma""")
    h.load_file("stdrun.hoc")
    h.soma.L = 5.6419
    h.soma.diam = 5.6419
    h.soma.insert("hh")
    ic = h.IClamp(h.soma(0.5))
    ic.delay = 0.5
    ic.dur = 0.1
    ic.amp = 0.3

    v = h.Vector()
    v.record(h.soma(0.5)._ref_v, sec=h.soma)
    tv = h.Vector()
    tv.record(h._ref_t, sec=h.soma)
    nc = h.NetCon(h.soma(0.5)._ref_v, None, sec=h.soma)
    spikestime = h.Vector()
    nc.record(spikestime)
    h.run()

    simulation_spikes = spikestime.to_python()

    assert np.allclose(simulation_spikes, ref_spikes)


def test_nrntest_test_2():
    h = hoc.HocObject()

    h(
        """
    create axon, soma, dend[3]
    connect axon(0), soma(0)
    for i=0, 2 connect dend[i](0), soma(1)
    axon { nseg = 5  L=100 diam = 1 insert hh }
    forsec "dend" { nseg = 3 L=50 diam = 2 insert pas e_pas = -65 }
    soma { L=10 diam=10 insert hh }
    """
    )

    assert h.axon.name() == "axon"
    assert str(h.dend) == "dend[?]"
    assert str(h.dend[1]) == "dend[1]"
    assert str(h.dend[1].name()) == "dend[1]"

    def e(stmt):
        err = 0
        try:
            exec(stmt, globals(), locals())
        except Exception:
            assert "not assignable" in str(sys.exc_info()[1])
            err = 1
        assert err == 1

    e("h.axon = 1")
    e("h.dend[1] = 1")

    assert str(h) == "<TopLevelHocInterpreter>"
    assert str(h.axon) == "axon"
    assert str(h.axon.name()) == "axon"
    assert str(h.axon(0.5)) == "axon(0.5)"
    assert str(h.axon(0.5).hh) == "hh"
    assert str(h.axon(0.5).hh.name()) == "hh"
    assert h.axon(0.5).hh.gnabar == 0.12
    assert h.axon(0.5) in h.axon
    assert h.axon(0.5) not in h.soma


def test_newobject_err_recover():
    err = 0
    try:
        nc = h.NetCon(5, 12)  # raises error
    except:
        err = 1
    assert err == 1
    h.finitialize()  # succeeds without seg fault


def test_push_section():
    h("""create hCable1, hCable2""")
    h.push_section("hCable1")
    assert h.secname() == "hCable1"
    h.pop_section()
    h.push_section("hCable2")
    assert h.secname() == "hCable2"
    h.pop_section()
    h.delete_section(sec=h.hCable1)
    h.delete_section(sec=h.hCable2)

    sections = [h.Section(name="pCable%d" % i) for i in range(2)]
    sections.append(h.Section())  # Anonymous in the past
    for sec in sections:
        name_in_hoc = h.secname(sec=sec)
        h.push_section(name_in_hoc)
        assert h.secname() == name_in_hoc
        h.pop_section()
    s = sections[-1]
    h.delete_section(sec=s)  # but not yet freed (though the name is now "")
    # not [no longer] a section pointer
    expect_err(
        "s.hoc_internal_name()"
    )  # this is what is generating the error in the next statement.
    expect_err('h.push_section(int(s.hoc_internal_name().replace("__nrnsec_", ""), 0))')

    # not a sectionname
    expect_hocerr(h.push_section, ("not_a_sectionname",))


def test_nonvint_block_supervisor():
    # making sure we can import
    from neuron import nonvint_block_supervisor

    # and can still use threads (no registered callbacks)
    pc = h.ParallelContext()
    pc.nthread(2)
    h.finitialize()
    # but if there is a callback it will fail with threads
    nonvint_block_supervisor.register([None for _ in range(11)])
    expect_hocerr(h.finitialize, (-65,))
    # but can be cleared
    nonvint_block_supervisor.clear()
    h.finitialize()
    pc.nthread(1)


# Before this test was introduced, HOC Object deletion was deferred
# when a Python HocObject was destoyed.
# Now deferred deletion is avoided for this case.
h(
    """
begintemplate Cell
    objref all
    create soma[1], dend[1]
    proc init() {
        all = new SectionList()
        soma {all.append()}
        dend {all.append()}
    }
endtemplate Cell
"""
)


def test_HocObject_no_deferred_unref():
    # previous tests should really destroy the model they created
    for sec in h.allsec():
        h.delete_section(sec=sec)

    sl = h.SectionList()
    cell = h.Cell()
    assert len(list(cell.all)) == 2

    del cell

    # When deferred deletion was in effect, following for loop  would segfault
    # because actual deletion of the HOC Cell object would not occur til the next
    # h.functioncall(). I.e. the first call to sl.append below would destroy
    # the HOC Cell which would destroy the two Sections and invalidate the
    # iterator causing a segfault.
    for sec in h.allsec():
        print(sec)
        sl.append(sec=sec)
    assert len([s for s in sl]) == 0


def test_deleted_sec():
    for sec in h.allsec():
        h.delete_section(sec=sec)
    s = h.Section()
    s.insert("hh")
    seg = s(0.5)
    ic = h.IClamp(seg)
    mech = seg.hh
    rvlist = [rv for rv in mech]
    vref = seg._ref_v
    gnabarref = mech._ref_gnabar
    sec_methods_ok = set([s.cell, s.name, s.hname, s.same, s.hoc_internal_name])
    sec_methods_chk = set(
        [
            getattr(s, n)
            for n in dir(s)
            if "__" not in n
            and type(getattr(s, n)) == type(s.x3d)
            and getattr(s, n) not in sec_methods_ok
        ]
    )
    seg_methods_chk = set(
        [
            getattr(seg, n)
            for n in dir(seg)
            if "__" not in n and type(getattr(seg, n)) == type(seg.area)
        ]
    )
    mech_methods_chk = set(
        [
            getattr(mech, n)
            for n in dir(mech)
            if "__" not in n and type(getattr(mech, n)) == type(mech.segment)
        ]
    )
    rv_methods_chk = set(
        [
            getattr(rvlist[0], n)
            for n in dir(rvlist[0])
            if "__" not in n and type(getattr(rvlist[0], n)) == type(rvlist[0].name)
        ]
    )
    rv_methods_chk.remove(rvlist[0].name)
    h.delete_section(sec=s)

    for methods in [sec_methods_chk, seg_methods_chk, mech_methods_chk, rv_methods_chk]:
        for m in methods:
            # Most would fail because of no args, but expect a check
            # for valid section first.
            words = str(m).split()
            print("m is " + words[4] + "." + words[2])
            expect_err("m()")

    assert str(s) == "<deleted section>"
    assert str(seg) == "<segment of deleted section>"
    assert str(mech) == "<mechanism of deleted section>"
    expect_err("h.sin(0.0, sec=s)")
    expect_err("print(s.L)")
    expect_err("s.L = 100.")
    expect_err("s.nseg")
    expect_err("s.nseg = 5")
    expect_err("print(s.v)")
    expect_err("print(s.orientation())")
    expect_err("s.insert('pas')")
    expect_err("s.insert(h.pas)")
    expect_err("s(.5)")
    expect_err("s(.5).v")
    expect_err("seg.v")
    expect_err("seg._ref_v")
    expect_err("s(.5).v = -65.")
    expect_err("seg.v = -65.")
    expect_err("seg.gnabar_hh")
    expect_err("mech.gnabar")
    expect_err("seg.gnabar_hh = .1")
    expect_err("mech.gnabar = .1")
    expect_err("rvlist[0][0]")
    expect_err("rvlist[0][0] = .1")
    expect_err("for sg in s: pass")
    expect_err("for m in seg: pass")
    expect_err("for r in mech: pass")

    assert s == s
    # See https://github.com/neuronsimulator/nrn/issues/1343
    if sys.version_info >= (3, 8):
        expect_err("dir(s)")
        expect_err("dir(seg)")
        expect_err("dir(mech)")
    assert type(dir(rvlist[0])) == list
    expect_err("help(s)")
    expect_err("help(seg)")
    expect_err("help(mech)")
    help(rvlist[0])  # not an error since it does not access s

    dend = h.Section()
    expect_err("dend.connect(s)")
    expect_err("dend.connect(seg)")

    # Note vref and gnabarref if used may cause segfault or other indeterminate result

    assert ic.get_segment() == None
    assert ic.has_loc() == 0.0
    expect_err("ic.get_loc()")
    expect_err("ic.amp")
    expect_err("ic.amp = .001")
    ic.loc(dend(0.5))
    assert ic.get_segment() == dend(0.5)

    imp = h.Impedance()
    expect_err("imp.loc(seg)")
    expect_err("h.distance(0, seg)")
    expect_hocerr(imp.loc, (seg,))
    expect_hocerr(h.distance, (0, seg))

    del ic, imp, dend
    locals()


def test_disconnect():
    print("test_disconnect")
    for sec in h.allsec():
        h.delete_section(sec=sec)
    h.topology()
    n = 5

    def setup(n):
        sections = [h.Section(name="s%d" % i) for i in range(n)]
        for i, sec in enumerate(sections[1:]):
            sec.connect(sections[i])
        return sections

    sl = setup(n)

    def chk(sections, i):
        print(sections, i)
        h.topology()
        x = len(sections[0].wholetree())
        assert x == i

    chk(sl, n)

    h.disconnect(sec=sl[2])
    chk(sl, 2)

    sl = setup(n)
    sl[2].disconnect()
    chk(sl, 2)

    sl = setup(n)
    expect_err("h.disconnect(sl[2])")
    expect_err("h.delete_section(sl[2])")

    del sl
    locals()


def test_py_alltoall_dict_err():
    pc = h.ParallelContext()
    src = {i: (100 + i) for i in range(2)}
    expect_hocerr(pc.py_alltoall, src, ("hocobj_call error",))


def test_nosection():
    expect_err("h.IClamp(.5)")
    expect_err("h.IClamp(5)")
    s = h.Section()
    expect_err("h.IClamp(5)")
    del s
    locals()


def test_nrn_mallinfo():
    # figure out if ASan was enabled, see comment in unit_test.cpp
    cmake_args = h.nrnversion(6)
    if re.search("'NRN_SANITIZERS=[a-z,]*address[a-z,]*'", cmake_args):
        print(
            "Skipping nrn_mallinfo checks because ASan was enabled ({})".format(
                cmake_args
            )
        )
        return
    assert h.nrn_mallinfo(0) > 0


if __name__ == "__main__":
    set_quiet(False)
    test_soma()
    test_simple_sim()
    test_deleted_sec()
    test_disconnect()
    h.topology()
    h.allobjects()
    test_nosection()
    test_nrn_mallinfo()



from .testutils import compare_data, tol


def test_include_flux(neuron_instance):
    """Test that node.include_flux gives the correct concentrations for 1D.

    node.include_flux can take either a; integer, Python function or NEURON
    pointer. All three are tested here for 1D intercellular rxd.
    """

    h, rxd, data, save_path = neuron_instance
    sec = h.Section(name="sec")
    sec.L = 10
    sec.nseg = 11
    sec.diam = 5

    cyt = rxd.Region(h.allsec(), name="cyt", nrn_region="i")
    ip3 = rxd.Species(
        cyt, name="ip3", initial=lambda nd: 1000 if nd.segment == sec(0.3) else 0
    )

    def callbackfun():
        return 1000

    h.finitialize(-70)

    for nd in ip3.nodes(sec(0.1)):
        nd.include_flux(1000)

    for nd in ip3.nodes(sec(0.5)):
        nd.include_flux(callbackfun)

    for nd in ip3.nodes(sec(0.9)):
        nd.include_flux(sec(0.3)._ref_ip3i)

    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol
